# Portions copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations

import operator
from typing import cast

# pyre-fixme[21]: Could not find name `INTRINSIC_1` in `cinderx.compiler.opcodes`.
from .opcodes import find_op_idx, INTRINSIC_1
from .optimizer import PyLimits, safe_lshift, safe_mod, safe_multiply, safe_power

TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Callable

    from .pyassem import Block, Instruction, PyFlowGraph

    Handler = Callable[
        [
            "FlowGraphOptimizer",
            int,
            Instruction,
            Instruction | None,
            Instruction | None,
            Block,
        ],
        int | None,
    ]


PyCmp_LT = 0
PyCmp_LE = 1
PyCmp_EQ = 2
PyCmp_NE = 3
PyCmp_GT = 4
PyCmp_GE = 5
PyCmp_IN = 6
PyCmp_NOT_IN = 7
PyCmp_IS = 8
PyCmp_IS_NOT = 9
PyCmp_EXC_MATCH = 10

STACK_USE_GUIDELINE = 30
MIN_CONST_SEQUENCE_SIZE = 3


UNARY_OPS: dict[str, object] = {
    "UNARY_INVERT": lambda v: ~v,
    "UNARY_NEGATIVE": lambda v: -v,
    "UNARY_POSITIVE": lambda v: +v,
}

BINARY_OPS: dict[int, Callable[[object, object], object]] = {
    find_op_idx("NB_POWER"): lambda x, y: safe_power(x, y, PyLimits),
    find_op_idx("NB_MULTIPLY"): lambda x, y: safe_multiply(x, y, PyLimits),
    find_op_idx("NB_TRUE_DIVIDE"): lambda left, right: left / right,
    find_op_idx("NB_FLOOR_DIVIDE"): lambda left, right: left // right,
    find_op_idx("NB_REMAINDER"): lambda x, y: safe_mod(x, y, PyLimits),
    find_op_idx("NB_ADD"): lambda left, right: left + right,
    find_op_idx("NB_SUBTRACT"): lambda left, right: left - right,
    find_op_idx("NB_SUBSCR"): lambda left, right: left[right],
    find_op_idx("NB_LSHIFT"): lambda x, y: safe_lshift(x, y, PyLimits),
    find_op_idx("NB_RSHIFT"): lambda left, right: left >> right,
    find_op_idx("NB_AND"): lambda left, right: left & right,
    find_op_idx("NB_XOR"): lambda left, right: left ^ right,
    find_op_idx("NB_OR"): lambda left, right: left | right,
}

SWAPPABLE: set[str] = {"STORE_FAST", "STORE_FAST_MAYBE_NULL", "POP_TOP"}


class FlowGraphOptimizer:
    """Flow graph optimizer."""

    JUMP_ABS: str = "<INVALID JUMP OPCODE>"  # Set to different opcodes in 3.10 and 3.12

    def __init__(self, graph: PyFlowGraph) -> None:
        self.graph = graph

    def optimize_basic_block(self, block: Block) -> None:
        raise NotImplementedError()

    def set_to_nop(self, instr: Instruction) -> None:
        raise NotImplemented()

    def dispatch_instr(
        self,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        handler = self.handlers.get(instr.opname)
        if handler is not None:
            return handler(self, instr_index, instr, next_instr, target, block)

    def skip_nops(self, next_block: Block, lineno: int) -> bool:
        return False

    def clean_basic_block(self, block: Block, prev_lineno: int) -> bool:
        """Remove all NOPs from a function when legal."""
        new_instrs = []
        num_instrs = len(block.insts)
        for idx in range(num_instrs):
            instr = block.insts[idx]
            if instr.opname == "NOP":
                lineno = instr.lineno
                # Eliminate no-op if it doesn't have a line number
                if lineno < 0:
                    continue
                # or, if the previous instruction had the same line number.
                if prev_lineno == lineno:
                    continue
                # or, if the next instruction has same line number or no line number
                if idx < num_instrs - 1:
                    next_instr = block.insts[idx + 1]
                    next_lineno = next_instr.lineno
                    if next_lineno == lineno:
                        continue
                    elif next_lineno < 0:
                        next_instr.loc = instr.loc
                        continue
                else:
                    next_block = block.next
                    while next_block and len(next_block.insts) == 0:
                        next_block = next_block.next
                    # or if last instruction in BB and next BB has same line number
                    if next_block:
                        if lineno == next_block.insts[0].lineno or self.skip_nops(
                            next_block, lineno
                        ):
                            continue
            new_instrs.append(instr)
            prev_lineno = instr.lineno
        cleaned = len(block.insts) != len(new_instrs)
        block.insts = new_instrs
        return cleaned

    def jump_thread(
        self, block: Block, instr: Instruction, target: Instruction, opname: str
    ) -> int:
        raise NotImplementedError()

    def get_const_loading_instrs(
        self, block: Block, start: int, size: int
    ) -> list[Instruction] | None:
        raise NotImplementedError()

    def opt_jump_if_false_or_pop(
        self,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        assert target is not None
        if target.opname == "POP_JUMP_IF_FALSE":
            return instr_index + self.jump_thread(
                block, instr, target, "POP_JUMP_IF_FALSE"
            )
        elif target.opname in (self.JUMP_ABS, "JUMP_FORWARD", "JUMP_IF_FALSE_OR_POP"):
            return instr_index + self.jump_thread(
                block, instr, target, "JUMP_IF_FALSE_OR_POP"
            )
        elif target.opname in ("JUMP_IF_TRUE_OR_POP", "POP_JUMP_IF_TRUE"):
            if instr.lineno == target.lineno:
                target_block = instr.target
                assert target_block and target_block != target_block.next
                instr.opname = "POP_JUMP_IF_FALSE"
                instr.target = target_block.next
                return instr_index
            return instr_index + 1

    def opt_jump_if_true_or_pop(
        self,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        assert target is not None
        if target.opname == "POP_JUMP_IF_TRUE":
            return instr_index + self.jump_thread(
                block, instr, target, "POP_JUMP_IF_TRUE"
            )
        elif target.opname in (self.JUMP_ABS, "JUMP_FORWARD", "JUMP_IF_TRUE_OR_POP"):
            return instr_index + self.jump_thread(
                block, instr, target, "JUMP_IF_TRUE_OR_POP"
            )
        elif target.opname in ("JUMP_IF_FALSE_OR_POP", "POP_JUMP_IF_FALSE"):
            if instr.lineno == target.lineno:
                target_block = instr.target
                assert target_block and target_block != target_block.next
                instr.opname = "POP_JUMP_IF_TRUE"
                instr.target = target_block.next
                return instr_index
            return instr_index + 1

    def opt_pop_jump_if(
        self,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        assert target is not None
        if target.opname in (self.JUMP_ABS, "JUMP_FORWARD", "JUMP"):
            return instr_index + self.jump_thread(block, instr, target, instr.opname)

    def opt_jump(
        self,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        assert target is not None
        if target.opname in (self.JUMP_ABS, "JUMP_FORWARD"):
            return instr_index + self.jump_thread(block, instr, target, self.JUMP_ABS)

    def opt_for_iter(
        self,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        assert target is not None
        if target.opname == "JUMP_FORWARD":
            return instr_index + self.jump_thread(block, instr, target, "FOR_ITER")

    def opt_rot_n(
        self,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        if instr.ioparg < 2:
            self.set_to_nop(instr)
            return
        elif instr.ioparg == 2:
            instr.opname = "ROT_TWO"
        elif instr.ioparg == 3:
            instr.opname = "ROT_THREE"
        elif instr.ioparg == 4:
            instr.opname = "ROT_FOUR"
        if instr_index >= instr.ioparg - 1:
            self.fold_rotations(
                block.insts[instr_index - instr.ioparg + 1 : instr_index + 1],
                instr.ioparg,
            )

    def fold_rotations(self, instrs: list[Instruction], n: int) -> None:
        for instr in instrs:
            if instr.opname == "ROT_N":
                rot = instr.ioparg
            elif instr.opname == "ROT_FOUR":
                rot = 4
            elif instr.opname == "ROT_THREE":
                rot = 3
            elif instr.opname == "ROT_TWO":
                rot = 2
            else:
                return
            if rot != n:
                return
        for instr in instrs:
            self.set_to_nop(instr)

    def opt_load_const(
        self,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        # Remove LOAD_CONST const; conditional jump
        const = instr.oparg
        if next_instr is None:
            return

        if next_instr.opname in (
            "POP_JUMP_IF_FALSE",
            "POP_JUMP_IF_TRUE",
            "JUMP_IF_FALSE",
            "JUMP_IF_TRUE",
        ):
            is_true = bool(const)
            if (
                next_instr.opname == "POP_JUMP_IF_FALSE"
                or next_instr.opname == "POP_JUMP_IF_TRUE"
            ):
                self.set_to_nop(block.insts[instr_index])
            jump_if_true = (
                next_instr.opname == "POP_JUMP_IF_TRUE"
                or next_instr.opname == "JUMP_IF_TRUE"
            )
            if is_true == jump_if_true:
                next_instr.opname = self.JUMP_ABS
                block.has_fallthrough = False
            else:
                next_instr.target = None
                self.set_to_nop(next_instr)
        elif next_instr.opname in ("JUMP_IF_FALSE_OR_POP", "JUMP_IF_TRUE_OR_POP"):
            is_true = bool(const)
            jump_if_true = next_instr.opname == "JUMP_IF_TRUE_OR_POP"
            if is_true == jump_if_true:
                next_instr.opname = self.JUMP_ABS
                block.has_fallthrough = False
            else:
                self.set_to_nop(block.insts[instr_index])
                self.set_to_nop(next_instr)

    def fold_tuple_on_constants(
        self, instr_index: int, instr: Instruction, block: Block
    ) -> None:
        load_const_instrs = []
        for i in range(instr_index - instr.ioparg, instr_index):
            maybe_load_const = block.insts[i]
            if maybe_load_const.opname != "LOAD_CONST":
                return
            load_const_instrs.append(maybe_load_const)
        newconst = tuple(lc.oparg for lc in load_const_instrs)
        for lc in load_const_instrs:
            self.set_to_nop(lc)
        instr.opname = "LOAD_CONST"
        instr.oparg = newconst
        instr.ioparg = self.graph.convertArg("LOAD_CONST", newconst)

    def opt_return_value(
        self,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        block.insts = block.insts[: instr_index + 1]

    handlers: dict[str, Handler] = {
        "JUMP_IF_FALSE_OR_POP": opt_jump_if_false_or_pop,
        "JUMP_IF_TRUE_OR_POP": opt_jump_if_true_or_pop,
        "POP_JUMP_IF_FALSE": opt_pop_jump_if,
        "POP_JUMP_IF_TRUE": opt_pop_jump_if,
        "JUMP_FORWARD": opt_jump,
        "FOR_ITER": opt_for_iter,
        "ROT_N": opt_rot_n,
        "LOAD_CONST": opt_load_const,
        "RETURN_VALUE": opt_return_value,
    }


class FlowGraphOptimizer310(FlowGraphOptimizer):
    """Python 3.10-specifc optimizations."""

    def opt_build_tuple(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        if (
            next_instr
            and next_instr.opname == "UNPACK_SEQUENCE"
            and instr.ioparg == next_instr.ioparg
        ):
            if instr.ioparg == 1:
                self.set_to_nop(instr)
                self.set_to_nop(next_instr)
            elif instr.ioparg == 2:
                instr.opname = "ROT_TWO"
                self.set_to_nop(next_instr)
            elif instr.ioparg == 3:
                instr.opname = "ROT_THREE"
                next_instr.opname = "ROT_TWO"
            return
        if instr_index >= instr.ioparg:
            self.fold_tuple_on_constants(instr_index, instr, block)

    JUMP_ABS = "JUMP_ABSOLUTE"
    handlers: dict[str, Handler] = {
        **FlowGraphOptimizer.handlers,
        JUMP_ABS: FlowGraphOptimizer.opt_jump,
        "BUILD_TUPLE": opt_build_tuple,
    }

    def set_to_nop(self, instr: Instruction) -> None:
        instr.opname = "NOP"

    def jump_thread(
        self, block: Block, instr: Instruction, target: Instruction, opname: str
    ) -> int:
        """Attempt to eliminate jumps to jumps by updating inst to jump to
        target->i_target using the provided opcode. Return 0 if successful, 1 if
        not; this makes it easier for our callers to revisit the same
        instruction again only if we changed it."""
        assert instr.is_jump(self.graph.opcode)
        assert target.is_jump(self.graph.opcode)
        if instr.lineno == target.lineno and instr.target != target.target:
            instr.target = target.target
            instr.opname = opname
            return 0
        return 1

    def optimize_basic_block(self, block: Block) -> None:
        instr_index = 0

        while instr_index < len(block.insts):
            instr = block.insts[instr_index]

            target_instr: Instruction | None = None
            if instr.is_jump(self.graph.opcode):
                target = instr.target
                assert target is not None
                # Skip over empty basic blocks.
                while len(target.insts) == 0:
                    instr.target = target.next
                    target = instr.target
                    assert target is not None
                target_instr = target.insts[0]

            next_instr = (
                block.insts[instr_index + 1]
                if instr_index + 1 < len(block.insts)
                else None
            )

            new_index = self.dispatch_instr(
                instr_index, instr, next_instr, target_instr, block
            )
            instr_index = instr_index + 1 if new_index is None else new_index


LOAD_CONST_INSTRS = ("LOAD_CONST", "LOAD_SMALL_INT")


class FlowGraphOptimizer312(FlowGraphOptimizer):
    """Python 3.12-specifc optimizations."""

    JUMP_ABS = "JUMP"

    def set_to_nop(self, instr: Instruction) -> None:
        instr.set_to_nop()

    def optimize_basic_block(self, block: Block) -> None:
        instr_index = 0
        opt_instr: Instruction | None = None
        # The handler needs to be called with the original instr and index
        # because the instr_index may be updated after the first call to
        # dispatch_instr.
        while instr_index < len(block.insts):
            instr = block.insts[instr_index]
            target_instr: Instruction | None = None

            is_copy_of_load_const = (
                opt_instr is not None
                and opt_instr.opname == "LOAD_CONST"
                and instr.opname == "COPY"
                and instr.ioparg == 1
            )

            if not is_copy_of_load_const:
                opt_instr = instr
                if instr.is_jump(self.graph.opcode):
                    target = instr.target
                    assert target is not None
                    assert target.insts, f"{instr} {target.label} {target.bid}"
                    # Skip over empty basic blocks.
                    while len(target.insts) == 0:
                        instr.target = target.next
                        target = instr.target
                        assert target is not None
                    target_instr = target.insts[0]

            next_instr = (
                block.insts[instr_index + 1]
                if instr_index + 1 < len(block.insts)
                else None
            )

            assert opt_instr is not None
            new_index = self.dispatch_instr(
                instr_index, opt_instr, next_instr, target_instr, block
            )

            instr_index = instr_index + 1 if new_index is None else new_index

    def try_opt_return_const(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> bool:
        if next_instr and next_instr.opname == "RETURN_VALUE":
            next_instr.opname = "RETURN_CONST"
            next_instr.oparg = instr.oparg
            next_instr.ioparg = instr.ioparg
            block.insts[instr_index].set_to_nop()
            return True

        return False

    def opt_load_const_is(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        jmp_op = (
            block.insts[instr_index + 2] if instr_index + 2 < len(block.insts) else None
        )
        if (
            jmp_op is not None
            and jmp_op.opname in ("POP_JUMP_IF_FALSE", "POP_JUMP_IF_TRUE")
            and instr.oparg is None
        ):
            nextarg = next_instr.oparg == 1
            instr.set_to_nop()
            next_instr.set_to_nop()
            jmp_op.opname = (
                "POP_JUMP_IF_NOT_NONE"
                if nextarg ^ (jmp_op.opname == "POP_JUMP_IF_FALSE")
                else "POP_JUMP_IF_NONE"
            )
            return instr_index + 2

    def opt_load_const(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        assert isinstance(self, FlowGraphOptimizer312)
        if self.try_opt_return_const(instr_index, instr, next_instr, target, block):
            return
        if next_instr is not None and next_instr.opname == "IS_OP":
            return self.opt_load_const_is(instr_index, instr, next_instr, target, block)
        else:
            # The rest of the optimizations are common to 3.10 and 3.12
            return super().opt_load_const(instr_index, instr, next_instr, target, block)

    def opt_push_null(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        if next_instr is None:
            return

        if next_instr.opname == "LOAD_GLOBAL" and (next_instr.ioparg & 1) == 0:
            instr.set_to_nop()
            next_instr.oparg = (next_instr.oparg, 1)
            next_instr.ioparg |= 1

    def opt_build_tuple(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        if (
            next_instr
            and next_instr.opname == "UNPACK_SEQUENCE"
            and instr.ioparg == next_instr.ioparg
        ):
            if instr.ioparg == 1:
                instr.set_to_nop()
                next_instr.set_to_nop()
                return
            elif instr.ioparg == 2 or instr.ioparg == 3:
                instr.set_to_nop()
                next_instr.opname = "SWAP"
                return
        if instr_index >= instr.ioparg:
            self.fold_tuple_on_constants(instr_index, instr, block)

    def opt_swap(
        self,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        if instr.oparg == 1:
            instr.set_to_nop()
            return

        new_index = self.swaptimize(instr_index, block)
        self.apply_static_swaps(new_index or instr_index, block)
        return new_index

    def next_swappable_instruction(self, i: int, block: Block, lineno: int) -> int:
        while i + 1 < len(block.insts):
            i += 1
            inst = block.insts[i]
            if lineno >= 0 and inst.lineno != lineno:
                # Optimizing across this instruction could cause user-visible
                # changes in the names bound between line tracing events!
                return -1
            elif inst.opname == "NOP":
                continue
            elif inst.opname in SWAPPABLE:
                return i

            return -1

        return -1

    # Attempt to apply SWAPs statically by swapping *instructions* rather than
    # stack items. For example, we can replace SWAP(2), POP_TOP, STORE_FAST(42)
    # with the more efficient NOP, STORE_FAST(42), POP_TOP.
    def apply_static_swaps(self, instr_index: int, block: Block) -> None:
        # SWAPs are to our left, and potential swaperands are to our right:
        for i in range(instr_index, -1, -1):
            swap = block.insts[i]
            if swap.opname != "SWAP":
                if swap.opname == "NOP" or swap.opname in SWAPPABLE:
                    # Nope, but we know how to handle these. Keep looking:
                    continue
                # We can't reason about what this instruction does. Bail:
                return

            j = self.next_swappable_instruction(i, block, -1)
            if j < 0:
                return

            k = j
            lineno = block.insts[j].lineno
            for _i in range(swap.ioparg - 1, 0, -1):
                k = self.next_swappable_instruction(k, block, lineno)
                if k < 0:
                    return

            # The reordering is not safe if the two instructions to be swapped
            # store to the same location, or if any intervening instruction stores
            # to the same location as either of them.
            store_j = block.insts[j].stores_to
            store_k = block.insts[k].stores_to
            if store_j is not None or store_k is not None:
                if store_j == store_k:
                    return
                for idx in range(j + 1, k):
                    store_idx = block.insts[idx].stores_to
                    if store_idx is not None and (
                        store_idx == store_j or store_idx == store_k
                    ):
                        return

            swap.set_to_nop()
            temp = block.insts[j]
            block.insts[j] = block.insts[k]
            block.insts[k] = temp

    def swaptimize(self, instr_index: int, block: Block) -> int | None:
        """Replace an arbitrary run of SWAPs and NOPs with an optimal one that has the
        same effect."""

        # Find the length of the current sequence of SWAPs and NOPs, and record the
        # maximum depth of the stack manipulations:
        instructions = block.insts
        depth = instructions[instr_index].ioparg
        more = False
        cnt = 0
        for cnt in range(instr_index + 1, len(instructions)):
            opname = instructions[cnt].opname
            if opname == "SWAP":
                depth = max(depth, instructions[cnt].ioparg)
                more = True
            elif opname != "NOP":
                break

        if not more:
            return None

        # Create an array with elements {0, 1, 2, ..., depth - 1}:
        stack = [i for i in range(depth)]
        # Simulate the combined effect of these instructions by "running" them on
        # our "stack":
        for i in range(instr_index, cnt):
            if block.insts[i].opname == "SWAP":
                oparg = instructions[i].ioparg
                top = stack[0]
                #  SWAPs are 1-indexed:
                stack[0] = stack[oparg - 1]
                stack[oparg - 1] = top

        ## Now we can begin! Our approach here is based on a solution to a closely
        ## related problem (https://cs.stackexchange.com/a/13938). It's easiest to
        ## think of this algorithm as determining the steps needed to efficiently
        ## "un-shuffle" our stack. By performing the moves in *reverse* order,
        ## though, we can efficiently *shuffle* it! For this reason, we will be
        ## replacing instructions starting from the *end* of the run. Since the
        ## solution is optimal, we don't need to worry about running out of space:
        current = cnt - instr_index - 1
        VISITED = -1
        for i in range(depth):
            if stack[i] == VISITED or stack[i] == i:
                continue

            # Okay, we've found an item that hasn't been visited. It forms a cycle
            # with other items; traversing the cycle and swapping each item with
            # the next will put them all in the correct place. The weird
            # loop-and-a-half is necessary to insert 0 into every cycle, since we
            # can only swap from that position:
            j = i
            while True:
                # Skip the actual swap if our item is zero, since swapping the top
                # item with itself is pointless:
                if j:
                    # SWAPs are 1-indexed:
                    instructions[current + instr_index].opname = "SWAP"
                    instructions[current + instr_index].ioparg = j + 1
                    current -= 1

                if stack[j] == VISITED:
                    # Completed the cycle:
                    assert j == i
                    break

                next_j = stack[j]
                stack[j] = VISITED
                j = next_j

        while 0 <= current:
            instructions[current + instr_index].set_to_nop()
            current -= 1

        return cnt

    def jump_thread(
        self, block: Block, instr: Instruction, target: Instruction, opname: str
    ) -> int:
        """Attempt to eliminate jumps to jumps by updating inst to jump to
        target->i_target using the provided opcode. Return 0 if successful, 1 if
        not; this makes it easier for our callers to revisit the same
        instruction again only if we changed it."""
        assert instr.is_jump(self.graph.opcode)
        assert target.is_jump(self.graph.opcode)
        if (
            instr.lineno == target.lineno or target.lineno == -1
        ) and instr.target != target.target:
            instr.target = target.target
            instr.opname = opname
            return 0
        return 1

    handlers: dict[str, Handler] = {
        **FlowGraphOptimizer.handlers,
        JUMP_ABS: FlowGraphOptimizer.opt_jump,
        "LOAD_CONST": opt_load_const,
        "PUSH_NULL": opt_push_null,
        "BUILD_TUPLE": opt_build_tuple,
        "SWAP": cast("Handler", opt_swap),
        "POP_JUMP_IF_NONE": FlowGraphOptimizer.opt_pop_jump_if,
        "POP_JUMP_IF_NOT_NONE": FlowGraphOptimizer.opt_pop_jump_if,
    }


def is_small_int(const: object) -> bool:
    return type(const) is int and const >= 0 and const < 256


class BaseFlowGraphOptimizer314(FlowGraphOptimizer312):
    def skip_nops(self, next_block: Block, lineno: int) -> bool:
        next_lineno = -1
        for next_instr in next_block.insts:
            # pyre-ignore[16]: no lineno
            if next_instr.opname == "NOP" and next_instr.loc.lineno < 0:
                # Skip over NOPs without a location, they will be removed
                continue
            # pyre-ignore[16]: no lineno
            next_lineno = next_instr.loc.lineno
        return lineno == next_lineno

    def opt_jump_if(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        assert target is not None
        if target.opname in ("JUMP", instr.opname):
            return instr_index + self.jump_thread(block, instr, target, instr.opname)
        elif target.opname == (
            "JUMP_IF_FALSE" if instr.opname == "JUMP_IF_TRUE" else "JUMP_IF_TRUE"
        ):
            # No need to check for loops here, a block's b_next
            # cannot point to itself.
            assert instr.target is not None
            instr.target = instr.target.next
            return instr_index - 1

    def get_const_loading_instrs(
        self, block: Block, start: int, size: int
    ) -> list[Instruction] | None:
        """Return a list of instructions that load the first `size` constants
        starting at `start`. Returns None if we don't have size constants."""
        const_loading_instrs = []
        if not size:
            return const_loading_instrs
        i = start
        while i >= 0:
            instr = block.insts[i]
            if instr.opname in LOAD_CONST_INSTRS:
                const_loading_instrs.append(instr)
                if len(const_loading_instrs) == size:
                    const_loading_instrs.reverse()
                    return const_loading_instrs
            elif instr.opname != "NOP":
                return
            i -= 1

    def optimize_lists_and_sets(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        assert isinstance(self, FlowGraphOptimizer314)
        contains_or_iter = next_instr is not None and (
            next_instr.opname == "GET_ITER" or next_instr.opname == "CONTAINS_OP"
        )
        seq_size = instr.oparg
        assert isinstance(seq_size, int)
        if seq_size > STACK_USE_GUIDELINE or (
            seq_size < MIN_CONST_SEQUENCE_SIZE and not contains_or_iter
        ):
            return

        const_loading_instrs = self.get_const_loading_instrs(
            block, instr_index - 1, seq_size
        )
        if const_loading_instrs is None:
            # If we're doing contains/iterating over a sequence we
            # know nothing will need to mutate it and a tuple is a
            # suitable container.
            if contains_or_iter and instr.opname == "BUILD_LIST":
                instr.opname = "BUILD_TUPLE"
            return

        newconst = tuple(i.oparg for i in const_loading_instrs)
        if instr.opname == "BUILD_SET":
            newconst = frozenset(newconst)

        index = self.graph.convertArg("LOAD_CONST", newconst)
        for lc in const_loading_instrs:
            lc.set_to_nop_no_loc()

        if contains_or_iter:
            instr.opname = "LOAD_CONST"
            instr.ioparg = instr.oparg = index
        else:
            assert instr_index >= 2
            assert instr.opname == "BUILD_LIST" or instr.opname == "BUILD_SET"

            block.insts[instr_index - 2].loc = instr.loc
            block.insts[instr_index - 2].opname = instr.opname
            block.insts[instr_index - 2].oparg = block.insts[instr_index - 2].ioparg = 0

            block.insts[instr_index - 1].opname = "LOAD_CONST"
            block.insts[instr_index - 1].ioparg = block.insts[instr_index - 1].oparg = (
                index
            )

            instr.opname = (
                "LIST_EXTEND" if instr.opname == "BUILD_LIST" else "SET_UPDATE"
            )
            instr.oparg = instr.ioparg = 1

    def fold_tuple_on_constants(
        self, instr_index: int, instr: Instruction, block: Block
    ) -> None:
        load_const_instrs = self.get_const_loading_instrs(
            block, instr_index - 1, instr.ioparg
        )
        if load_const_instrs is None:
            return
        newconst = tuple(lc.oparg for lc in load_const_instrs)
        for lc in load_const_instrs:
            lc.set_to_nop_no_loc()
        instr.opname = "LOAD_CONST"
        instr.oparg = newconst
        instr.ioparg = self.graph.convertArg("LOAD_CONST", newconst)

    def optimize_compare_op(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        if next_instr and next_instr.opname == "TO_BOOL":
            next_instr.opname = "COMPARE_OP"
            next_instr.oparg = next_instr.ioparg = instr.ioparg | 16
            instr.set_to_nop()

    def optimize_load_global(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        if next_instr is not None and next_instr.opname == "PUSH_NULL":
            instr.oparg = (instr.oparg, 1)
            instr.ioparg |= 1
            next_instr.set_to_nop()

    def opt_jump_no_interrupt(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        assert target is not None
        if target.opname == "JUMP":
            return instr_index + self.jump_thread(block, instr, target, "JUMP")
        elif target.opname == "JUMP_NO_INTERRUPT":
            return instr_index + self.jump_thread(
                block, instr, target, "JUMP_NO_INTERRUPT"
            )

    def jump_thread(
        self, block: Block, instr: Instruction, target: Instruction, opname: str
    ) -> int:
        """Attempt to eliminate jumps to jumps by updating inst to jump to
        target->i_target using the provided opcode. Return 0 if successful, 1 if
        not; this makes it easier for our callers to revisit the same
        instruction again only if we changed it."""
        assert instr.is_jump(self.graph.opcode)
        assert target.is_jump(self.graph.opcode)
        if instr.target != target.target:
            instr.set_to_nop()
            assert block.insts[-1] == instr
            block.append_instr(opname, 0, target=target.target, loc=target.loc)

        return 1

    def opt_store_fast(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        if not next_instr:
            return
        # Remove the store fast instruction if it is storing the same local as the
        # next instruction
        if (
            next_instr.opname == "STORE_FAST"
            and next_instr.ioparg == instr.ioparg
            # pyre-ignore[16]: `Instruction` has no attribute `loc`.
            and next_instr.loc.lineno == instr.loc.lineno
        ):
            instr.opname = "POP_TOP"
            instr.oparg = instr.ioparg = 0

    def optimize_contains_is_op(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        if next_instr is None:
            return
        if next_instr.opname == "TO_BOOL":
            next_instr.opname = instr.opname
            next_instr.oparg = instr.oparg
            next_instr.ioparg = instr.ioparg
            instr.set_to_nop()
        elif next_instr.opname == "UNARY_NOT":
            next_instr.opname = instr.opname
            next_instr.oparg = instr.oparg
            next_instr.ioparg = instr.ioparg ^ 1
            instr.set_to_nop()

    def make_load_const(self, instr: Instruction, const: object) -> None:
        if is_small_int(const):
            assert isinstance(const, int)
            instr.opname = "LOAD_SMALL_INT"
            instr.ioparg = instr.oparg = const
        else:
            instr.opname = "LOAD_CONST"
            instr.oparg = const
            instr.ioparg = self.graph.convertArg("LOAD_CONST", const)

    def optimize_binary_op(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        assert isinstance(self, FlowGraphOptimizer314)
        consts = self.get_const_loading_instrs(block, instr_index - 1, 2)
        if consts is None:
            return
        lhs = consts[0].oparg
        rhs = consts[1].oparg

        op = BINARY_OPS.get(instr.ioparg)
        if op is None:
            return

        try:
            res = op(lhs, rhs)
        except (ArithmeticError, TypeError, ValueError, IndexError):
            return

        consts[0].set_to_nop_no_loc()
        consts[1].set_to_nop_no_loc()
        self.make_load_const(instr, res)

    def optimize_one_unary(
        self,
        instr_index: int,
        instr: Instruction,
        block: Block,
        op: Callable[[object], object],
    ) -> object:
        consts = self.get_const_loading_instrs(block, instr_index - 1, 1)
        if consts is None:
            return

        try:
            res = op(consts[0].oparg)
        except (ArithmeticError, TypeError, ValueError, IndexError):
            return

        consts[0].set_to_nop_no_loc()
        self.make_load_const(instr, res)

    def optimize_unary_invert(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        assert isinstance(self, FlowGraphOptimizer314)
        self.optimize_one_unary(instr_index, instr, block, operator.inv)

    def optimize_unary_negative(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        assert isinstance(self, FlowGraphOptimizer314)
        self.optimize_one_unary(instr_index, instr, block, operator.neg)

    def optimize_unary_not(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        if next_instr is not None and next_instr.opname == "TO_BOOL":
            instr.set_to_nop()
            next_instr.opname = "UNARY_NOT"
            return
        if next_instr is not None and next_instr.opname == "UNARY_NOT":
            instr.set_to_nop()
            next_instr.set_to_nop()
            return

        assert isinstance(self, FlowGraphOptimizer314)
        self.optimize_one_unary(instr_index, instr, block, operator.not_)

    def fold_constrant_intrinsic_list_to_tuple(
        self, block: Block, instr_index: int
    ) -> None:
        consts_found = 0
        expect_append = True
        for i in range(instr_index - 1, -1, -1):
            instr = block.insts[i]
            opcode = instr.opname
            oparg = instr.oparg
            if opcode == "NOP":
                continue

            if opcode == "BUILD_LIST" and oparg == 0:
                if not expect_append:
                    # Not a start sequence
                    return

                # Sequence start, we are done.
                consts = []
                if opcode == "BUILD_LIST" and oparg == 0:
                    for newpos in range(instr_index - 1, i - 1, -1):
                        instr = block.insts[newpos]
                        if instr.opname in LOAD_CONST_INSTRS:
                            const = instr.oparg
                            consts.append(const)
                        instr.set_to_nop_no_loc()

                consts.reverse()
                self.make_load_const(block.insts[instr_index], tuple(consts))
                return

            if expect_append:
                if opcode != "LIST_APPEND" or oparg != 1:
                    return
            elif opcode not in LOAD_CONST_INSTRS:
                return
            consts_found += 1
            expect_append = not expect_append

    def optimize_call_instrinsic_1(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        assert isinstance(self, FlowGraphOptimizer314)
        # pyre-fixme[16]: Module `opcodes` has no attribute `INTRINSIC_1`.
        intrins = INTRINSIC_1[instr.ioparg]
        if intrins == "INTRINSIC_LIST_TO_TUPLE":
            if next_instr is not None and next_instr.opname == "GET_ITER":
                instr.set_to_nop()
            else:
                self.fold_constrant_intrinsic_list_to_tuple(block, instr_index)
        if intrins == "INTRINSIC_UNARY_POSITIVE":
            self.optimize_one_unary(instr_index, instr, block, operator.pos)

    def optimize_swap(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        if instr.oparg == 1:
            instr.set_to_nop()

    handlers: dict[str, Handler] = {
        **FlowGraphOptimizer312.handlers,
        "JUMP_IF_FALSE": opt_jump_if,
        "JUMP_IF_TRUE": opt_jump_if,
        "BUILD_LIST": optimize_lists_and_sets,
        "BUILD_SET": optimize_lists_and_sets,
        "COMPARE_OP": optimize_compare_op,
        "CONTAINS_OP": optimize_contains_is_op,
        "IS_OP": optimize_contains_is_op,
        "LOAD_GLOBAL": optimize_load_global,
        "JUMP_NO_INTERRUPT": opt_jump_no_interrupt,
        "STORE_FAST": opt_store_fast,
        "BINARY_OP": optimize_binary_op,
        "UNARY_INVERT": optimize_unary_invert,
        "UNARY_NEGATIVE": optimize_unary_negative,
        "UNARY_NOT": optimize_unary_not,
        "CALL_INTRINSIC_1": optimize_call_instrinsic_1,
        "SWAP": optimize_swap,
    }
    del handlers["PUSH_NULL"]
    del handlers["LOAD_CONST"]


class FlowGraphOptimizer314(BaseFlowGraphOptimizer314):
    def optimize_basic_block(self, block: Block) -> None:
        super().optimize_basic_block(block)
        i = 0
        while i < len(block.insts):
            inst = block.insts[i]
            if inst.opname == "SWAP":
                new_i = self.swaptimize(i, block)
                self.apply_static_swaps(new_i or i, block)
                if new_i is not None:
                    i = new_i
            i += 1


class FlowGraphConstOptimizer314(BaseFlowGraphOptimizer314):
    def opt_load_const(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction | None,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        assert isinstance(self, FlowGraphConstOptimizer314)
        if instr.opname == "LOAD_CONST" and is_small_int(instr.oparg):
            assert isinstance(instr.oparg, int)
            instr.ioparg = instr.oparg
            instr.opname = "LOAD_SMALL_INT"

        if next_instr is None:
            return

        if next_instr.opname == "TO_BOOL":
            val = bool(instr.oparg)
            index = self.graph.convertArg("LOAD_CONST", val)
            instr.set_to_nop_no_loc()
            next_instr.opname = "LOAD_CONST"
            next_instr.oparg = val
            next_instr.ioparg = index
        elif next_instr.opname == "IS_OP":
            return self.opt_load_const_is(instr_index, instr, next_instr, target, block)
        else:
            # The rest of the optimizations are common to 3.10 and 3.12
            return FlowGraphOptimizer.opt_load_const(
                self, instr_index, instr, next_instr, target, block
            )

    def opt_load_const_is(
        self: FlowGraphOptimizer,
        instr_index: int,
        instr: Instruction,
        next_instr: Instruction,
        target: Instruction | None,
        block: Block,
    ) -> int | None:
        jmp_op = (
            block.insts[instr_index + 2] if instr_index + 2 < len(block.insts) else None
        )
        if jmp_op is not None and jmp_op.opname == "TO_BOOL":
            jmp_op.set_to_nop()
            if instr_index + 3 >= len(block.insts):
                return
            jmp_op = block.insts[instr_index + 3]

        if (
            jmp_op is not None
            and jmp_op.opname in ("POP_JUMP_IF_FALSE", "POP_JUMP_IF_TRUE")
            and instr.oparg is None
        ):
            nextarg = next_instr.oparg == 1
            instr.set_to_nop()
            next_instr.set_to_nop()
            jmp_op.opname = (
                "POP_JUMP_IF_NOT_NONE"
                if nextarg ^ (jmp_op.opname == "POP_JUMP_IF_FALSE")
                else "POP_JUMP_IF_NONE"
            )
            return instr_index + 2

    handlers: dict[str, Handler] = {
        "LOAD_CONST": opt_load_const,
        "LOAD_SMALL_INT": opt_load_const,
    }



# Copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict

from __future__ import annotations

import ast
import sys
from typing import Callable, Mapping

COMPARE_OPS: Mapping[type[ast.cmpop], Callable[[object, object], bool]] = {
    ast.Gt: lambda a, b: a > b,
    ast.GtE: lambda a, b: a >= b,
    ast.Lt: lambda a, b: a < b,
    ast.LtE: lambda a, b: a <= b,
    ast.Eq: lambda a, b: a == b,
    ast.NotEq: lambda a, b: a != b,
    ast.In: lambda a, b: a in b,
    ast.Is: lambda a, b: a is b,
    ast.IsNot: lambda a, b: a is not b,
}


def make_qualname(parent_qualname: str | None, name: str) -> str:
    if parent_qualname is None:
        return name
    return f"{parent_qualname}.{name}"


def _is_sys_hexversion_attr_load(node: ast.expr) -> bool:
    if isinstance(node, ast.Attribute):
        container = node.value
        if (
            isinstance(container, ast.Name)
            and container.id == "sys"
            and isinstance(node.ctx, ast.Load)
            and node.attr == "hexversion"
        ):
            return True
    return False


def _get_const_int(node: ast.expr) -> int | None:
    if isinstance(node, ast.Constant):
        value = node.value
        return value if isinstance(value, int) else None


def sys_hexversion_check(
    node: ast.If,
) -> bool | None:
    """
    A helper function, the result of this is used to determine whether
    we need to skip visiting dead code gated by sys.hexversion checks.
    """
    test_node = node.test
    if isinstance(test_node, ast.Compare):
        if len(test_node.comparators) != 1:
            return None

        assert len(test_node.ops) == 1

        left = test_node.left
        right = test_node.comparators[0]
        op = test_node.ops[0]

        if type(op) in (ast.In, ast.Is, ast.IsNot):
            return None

        if _is_sys_hexversion_attr_load(left):
            left_value = sys.hexversion
            right_value = _get_const_int(right)
        elif _is_sys_hexversion_attr_load(right):
            left_value = _get_const_int(left)
            right_value = sys.hexversion
        else:
            return None

        if left_value is None or right_value is None:
            return None

        return COMPARE_OPS[type(op)](left_value, right_value)



# Portions copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict

"""Parser for future statements"""

import ast

from .visitor import ASTVisitor


class FutureParser(ASTVisitor):
    features = (
        "nested_scopes",
        "generators",
        "division",
        "absolute_import",
        "with_statement",
        "print_function",
        "unicode_literals",
        "generator_stop",
        "barry_as_FLUFL",
        "annotations",
    )

    def __init__(self) -> None:
        super().__init__()
        self.valid: set[ast.ImportFrom] = set()
        self.found_features: set[str] = set()
        self.possible_docstring = True

    def visitModule(self, node: ast.Module) -> None:
        for s in node.body:
            if (
                self.possible_docstring
                and isinstance(s, ast.Expr)
                and isinstance(s.value, ast.Constant)
                and isinstance(s.value.value, str)
            ):
                self.possible_docstring = False
                continue
            # no docstring after first statement
            self.possible_docstring = False
            if not self.check_stmt(s):
                break

    def check_stmt(self, stmt: ast.stmt) -> bool:
        if isinstance(stmt, ast.ImportFrom) and stmt.module == "__future__":
            for alias in stmt.names:
                name = alias.name
                if name in self.features:
                    self.found_features.add(name)
                elif name == "braces":
                    raise SyntaxError("not a chance")
                else:
                    raise SyntaxError("future feature %s is not defined" % name)
            self.valid.add(stmt)
            return True
        return False

    def get_features(self) -> set[str]:
        """Return set of features enabled by future statements"""
        return self.found_features

    def get_valid(self) -> set[ast.ImportFrom]:
        """Return set of valid future statements"""
        return self.valid


class BadFutureParser(ASTVisitor):
    """Check for invalid future statements"""

    def __init__(self, valid_parser: FutureParser) -> None:
        super().__init__()
        self.valid_parser = valid_parser

    def visitImportFrom(self, node: ast.ImportFrom) -> None:
        if node.module != "__future__":
            return
        if node in self.valid_parser.get_valid():
            return
        raise SyntaxError(
            "from __future__ imports must occur at the beginning of the file"
        )


def find_futures(node: ast.AST) -> set[str]:
    p1 = FutureParser()
    p2 = BadFutureParser(p1)
    p1.visit(node)
    p2.visit(node)
    return p1.get_features()



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations

import ast
import builtins
import sys
from ast import (
    AST,
    AsyncFunctionDef,
    Call,
    ClassDef,
    expr,
    FunctionDef,
    ImportFrom,
    Name,
    NodeVisitor,
    stmt,
)
from typing import Any, cast, final, Mapping

from ..consts import CO_FUTURE_ANNOTATIONS
from ..pyassem import PyFlowGraph
from ..pycodegen import CinderCodeGenBase, CodeGenerator, find_futures, FOR_LOOP
from ..symbols import BaseSymbolVisitor, FunctionScope
from .common import FIXED_MODULES, lineinfo
from .feature_extractor import _IMPLICIT_GLOBALS, FeatureExtractor


def is_mutable(node: AST) -> bool:
    return isinstance(node, Name) and node.id in ("mutable")


class FindClassDef(NodeVisitor):
    def __init__(self) -> None:
        self.has_class = False

    def check(self, node: AST) -> None:
        self.generic_visit(node)

    def visit_ClassDef(self, node: ClassDef) -> None:
        for n in node.decorator_list:
            if is_mutable(n):
                # do not consider this class for freezing
                break
        else:
            self.has_class = True
        for body_stmt in node.body:
            # also consider inner classes
            self.visit(body_stmt)

    def visit_FunctionDef(self, node: FunctionDef) -> None:
        # do not go into func body
        pass

    def visit_AsyncFunctionDef(self, node: AsyncFunctionDef) -> None:
        # do not go into func body
        pass


class ForBodyHook(ast.stmt):
    def __init__(self, node: list[ast.stmt], target: ast.expr) -> None:
        self.body: list[ast.stmt] = node
        self.target = target
        lineinfo(self, target)


class TryFinallyHook(ast.stmt):
    def __init__(self, finally_body: list[ast.stmt]) -> None:
        self.finally_body = finally_body
        # List of (handler, builtin_name)
        self.handlers_to_restore: list[tuple[str, str]] = []
        lineinfo(self, finally_body[0] if finally_body else None)


class TryHandlerBodyHook(ast.stmt):
    def __init__(self, handler_body: list[ast.stmt], tracker_name: str) -> None:
        self.handler_body = handler_body
        self.tracker_name = tracker_name
        lineinfo(self, handler_body[0] if handler_body else None)


class TryBodyHook(ast.stmt):
    def __init__(self, body: list[ast.stmt]) -> None:
        self.body = body
        lineinfo(self, body[0] if body else None)
        self.trackers: list[str] = []


class StrictCodeGenBase(CinderCodeGenBase):
    flow_graph = PyFlowGraph
    class_list_name: str = "<classes>"

    def __init__(
        self,
        parent: CodeGenerator | None,
        node: AST,
        symbols: BaseSymbolVisitor,
        graph: PyFlowGraph,
        flags: int = 0,
        optimization_lvl: int = 0,
        builtins: dict[str, Any] = builtins.__dict__,
        future_flags: int | None = None,
        name: str | None = None,
    ) -> None:
        super().__init__(
            parent,
            node,
            symbols,
            graph,
            flags=flags,
            optimization_lvl=optimization_lvl,
            future_flags=future_flags,
            name=name,
        )
        self.has_class: bool = self.has_classDef(node)
        self.made_class_list = False
        self.builtins = builtins
        # record the lineno of the first toplevel statement
        # this will be the lineno for all extra generated code
        # in the beginning of the strict module
        self.first_body_node: ast.stmt | None = None
        if not parent and isinstance(node, ast.Module) and len(node.body) > 0:
            self.first_body_node = node.body[0]

        if parent and isinstance(parent, StrictCodeGenBase):
            self.feature_extractor: FeatureExtractor = parent.feature_extractor
        else:
            self.feature_extractor = FeatureExtractor(builtins, self.future_flags)
            self.feature_extractor.visit(node)

    @classmethod
    def make_code_gen(
        cls,
        module_name: str,
        tree: AST,
        filename: str,
        source: str | bytes | ast.Module | ast.Expression | ast.Interactive,
        flags: int,
        optimize: int,
        ast_optimizer_enabled: bool = True,
        builtins: dict[str, Any] = builtins.__dict__,
    ) -> StrictCodeGenBase:
        future_flags = find_futures(flags, tree)
        if ast_optimizer_enabled:
            tree = cls.optimize_tree(
                optimize, tree, bool(future_flags & CO_FUTURE_ANNOTATIONS)
            )
        s = cls._SymbolVisitor(future_flags)
        s.visit(tree)

        graph = cls.flow_graph(module_name, filename, s.scopes[tree])
        code_gen = cls(
            None,
            tree,
            s,
            graph,
            flags=flags,
            optimization_lvl=optimize,
            builtins=builtins,
            future_flags=future_flags,
        )
        code_gen._qual_name = module_name
        code_gen.visit(tree)
        return code_gen

    def is_functionScope(self) -> bool:
        """Whether the generator has any parent function scope"""
        scope = self.scope
        while scope is not None:
            if isinstance(scope, FunctionScope):
                return True
            scope = scope.parent
        return False

    def has_classDef(self, node: AST) -> bool:
        visitor = FindClassDef()
        visitor.check(node)
        return visitor.has_class

    def emit_load_fixed_methods(self) -> None:
        # load and store <strict-modules>
        self.emit("LOAD_NAME", "<fixed-modules>")
        self.emit("LOAD_CONST", "__strict__")
        self.emit_binary_subscr()
        self.emit_dup()
        self.emit("STORE_GLOBAL", "<strict-modules>")
        self.emit("LOAD_CONST", "freeze_type")
        self.emit_binary_subscr()
        self.emit("STORE_GLOBAL", "<freeze-type>")

    def strictPreVisitCall(self, node: Call) -> None:
        func = node.func
        if isinstance(func, ast.Name):
            # We don't currently allow aliasing or shadowing exec/eval
            # so this check is currently sufficient.
            if func.id == "exec" or func.id == "eval":
                # exec and eval don't take keyword args, so we only need to
                # inspect the normal arguments
                if len(node.args) < 2:
                    # we're implicitly capturing globals, make it explicit so
                    # that we get our globals dictionary.  exec/eval won't be
                    # able to mutate them though.  We also need to explicitly
                    # call locals() here because the behavior of exec/eval is
                    # to use globals as locals if it is explicitly supplied.
                    def call_function(line_node: AST, name: str) -> expr:
                        load = lineinfo(ast.Name(name, ast.Load()))
                        call = lineinfo(ast.Call(load, [], []))
                        return call

                    node.args.append(call_function(node.args[0], "globals"))
                    node.args.append(call_function(node.args[0], "locals"))

    def visitCall(self, node: Call) -> None:
        self.strictPreVisitCall(node)

        super().visitCall(node)

    def visitForBodyHook(self, node: ForBodyHook) -> None:
        self.visit_list(node.body)

    def strictPreVisitFor(self, node: ast.For) -> None:
        node.body = [ForBodyHook(node.body, node.target)]

    def strictPostVisitFor(self, node: ast.For) -> None:
        # And undo to keep any introspection happy
        node.body = cast(ForBodyHook, node.body[0]).body

    def visitFor(self, node: ast.For) -> None:
        self.strictPreVisitFor(node)
        super().visitFor(node)
        self.strictPostVisitFor(node)

    def make_function(
        self, name: str, body: list[stmt], location_node: ast.AST | None = None
    ) -> None:
        # pyre-fixme[20]: Argument `args` expected.
        func = lineinfo(ast.FunctionDef(), location_node)
        func.name = name
        func.decorator_list = []
        func.returns = None
        func.type_comment = ""
        func.body = body
        # pyre-fixme[20]: Argument `args` expected.
        args = lineinfo(ast.arguments())
        func.args = args

        args.kwonlyargs = []
        args.kw_defaults = []
        args.defaults = []
        args.args = []
        args.posonlyargs = []
        args.kwarg = None
        args.vararg = None

        scope = FunctionScope(name, self.scope.module, self.scope.klass)
        scope.parent = self.scope
        self.feature_extractor.scopes[func] = scope
        self.scopes[func] = scope

        self.visitFunctionDef(func)

    def emit_append_class_list(self) -> None:
        """
        Assumes the class to append is TOS1.
        Assumes `<classes>` is at TOS
        Pops that class and append to `<classes>`
        Do not Leave `<classes>` on the stack
        """
        self.emit_rotate_stack(2)
        self.emit("LIST_APPEND", 1)
        self.emit("POP_TOP")

    def emit_create_class_list(self) -> None:
        """create and store an empty list `<classes>`"""
        self.emit("BUILD_LIST", 0)
        op = "STORE_FAST" if self.is_functionScope() else "STORE_GLOBAL"
        self.emit(op, self.class_list_name)
        self.made_class_list = True

    def emit_freeze_class_list(self) -> None:
        """
        create a for loop that iterates on all of `<classes>`
        assign the value to `<class>`, and call
        freeze_type on `<class>`
        """

        start = self.newBlock()
        anchor = self.newBlock()
        after = self.newBlock()
        body = self.newBlock()

        self.push_loop(FOR_LOOP, start, after)
        self.emit_load_class_list()
        self.emit("GET_ITER")

        self.nextBlock(start)
        # for <class> in <classes>
        # we don't actually need to assign to <class>
        self.emit("FOR_ITER", anchor)
        self.nextBlock(body)
        self.emit("LOAD_GLOBAL", "<freeze-type>")
        # argument need to be top most
        self.emit_rotate_stack(2)
        self.emit_call_one_arg()
        # discard the result of call
        self.emit("POP_TOP")

        self.emitJump(start)
        self.nextBlock(anchor)
        self.emit_end_for()
        self.pop_loop()
        self.nextBlock(after)

    def emit_load_class_list(self) -> None:
        """load `<classes>` on top of stack"""
        op = "LOAD_FAST" if self.is_functionScope() else "LOAD_GLOBAL"
        self.emit(op, self.class_list_name)

    def find_immutability_flag(self, node: ClassDef) -> bool:
        old_size = len(node.decorator_list)
        node.decorator_list = [d for d in node.decorator_list if not is_mutable(d)]
        return old_size == len(node.decorator_list)

    def register_immutability(self, node: ClassDef, flag: bool) -> None:
        if self.has_class and flag:
            self.emit_dup()
            self.emit_load_class_list()
            self.emit_append_class_list()
        super().register_immutability(node, flag)

    def processBody(
        self,
        node: ast.FunctionDef | ast.AsyncFunctionDef | ast.Lambda,
        body: list[ast.stmt],
        gen: CodeGenerator,
    ) -> None:
        if (
            isinstance(node, (FunctionDef, AsyncFunctionDef))
            and isinstance(gen, StrictCodeGenBase)
            and gen.has_class
        ):
            # initialize the <classes> list
            if not gen.made_class_list:
                gen.emit_create_class_list()
            # create a try + finally structure where we freeze all classes
            # in the finally block
            gen.emit_try_finally(
                None,
                lambda: super(StrictCodeGenBase, self).processBody(node, body, gen),
                lambda: gen.emit_freeze_class_list(),
            )
        else:
            super().processBody(node, body, gen)

    def startModule(self) -> None:
        first_node = self.first_body_node
        if first_node:
            self.set_pos(first_node)
        self.emit_load_fixed_methods()
        if self.has_class and not self.made_class_list:
            self.emit_create_class_list()
        super().startModule()

    def emit_module_return(self, node: ast.Module) -> None:
        if self.has_class:
            self.emit_freeze_class_list()
        super().emit_module_return(node)

    def emit_import_fixed_modules(
        self, node: ImportFrom, mod_name: str, mod: Mapping[str, object]
    ) -> None:
        new_names = [n for n in node.names if n.name not in mod]
        if new_names:
            # We have names we don't know about, keep them around...
            new_node = ImportFrom(mod_name, new_names, node.level)
            super().visitImportFrom(new_node)

        # Load the module into TOS...
        self.emit("LOAD_NAME", "<fixed-modules>")
        self.emit("LOAD_CONST", mod_name)
        self.emit_binary_subscr()  # TOS = mod

        # Store all of the imported names from the module
        for _i, name in enumerate(node.names):
            var_name = name.name
            asname = name.asname or var_name
            value = mod.get(var_name)
            if value is not None:
                # duplicate TOS (mod)
                self.emit_dup()
                # var name
                self.emit("LOAD_CONST", var_name)
                self.emit_binary_subscr()
                self.emit("STORE_GLOBAL", asname)
        # remove TOS (mod)
        self.emit("POP_TOP")

    @final
    def visitImportFrom(self, node: ImportFrom) -> None:
        if node.level == 0 and node.module is not None and node.module in FIXED_MODULES:
            module_name = node.module
            assert module_name is not None
            mod = FIXED_MODULES[module_name]
            self.emit_import_fixed_modules(node, module_name, mod)
            return
        super().visitImportFrom(node)



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations


try:
    # pyre-ignore[21]: Undefined import
    from cinderx import freeze_type as cinder_freeze
except ImportError:
    cinder_freeze = None


try:
    from cinder import warn_on_inst_dict
except ImportError:
    warn_on_inst_dict = None


__all__ = [
    "freeze_type",
    "loose_slots",
    "strict_slots",
    "mutable",
    "extra_slot",
    "_mark_cached_property",
    "set_freeze_enabled",
]

TYPE_FREEZE_ENABLED = True


def set_freeze_enabled(flag: bool) -> bool:
    "returns old value"
    global TYPE_FREEZE_ENABLED
    old = TYPE_FREEZE_ENABLED
    TYPE_FREEZE_ENABLED = flag
    return old


def freeze_type(obj: type[object]) -> type[object]:
    if cinder_freeze is not None:
        if TYPE_FREEZE_ENABLED:
            cinder_freeze(obj)
        elif warn_on_inst_dict is not None:
            warn_on_inst_dict(obj)

    return obj


def loose_slots(obj: type[object]) -> type[object]:
    """Indicates that a type defined in a strict module should support assigning
    additional variables to __dict__ to support migration."""
    if warn_on_inst_dict is not None:
        warn_on_inst_dict(obj)
    return obj


def mutable(obj: type[object]) -> type[object]:
    """Marks a type defined in a strict module as supporting mutability"""
    return obj


def strict_slots(obj: type[object]) -> type[object]:
    """Marks a type defined in a strict module to get slots automatically
    and no __dict__ is created"""
    return obj


def extra_slot(obj: type[object], _name: str) -> type[object]:
    """mark `name` to be part of __slots__ in obj"""
    return obj


def _mark_cached_property(obj: object, is_async: bool, original_dec: object) -> object:
    return obj



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

try:
    from cinderx.compiler.strict.runtime import (
        _mark_cached_property,
        extra_slot,
        freeze_type,
        loose_slots,
        mutable,
        set_freeze_enabled,
        strict_slots,
    )
except ImportError:

    def _mark_cached_property(o: object) -> object:
        return o

    def extra_slot(cls: object, o: object) -> object:
        return cls

    def freeze_type(o: object) -> object:
        return o

    def loose_slots(o: object) -> object:
        return o

    def mutable(o: object) -> object:
        return o

    def set_freeze_enabled(o: object) -> object:
        return o

    def strict_slots(o: object) -> object:
        return o


allow_side_effects = object()



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-unsafe

# Generated via assign_opcode_numbers.py, do not edit.

# This is an addition to python/3.12/Lib/opcode.py, and is intended to be run
# via `exec` in generate_opcode_h.py with the globals dict obtained from
# running Lib/opcode.py.

# flake8: noqa


# Lib/opcode.py deletes these functions so we need to define them again here.
# We also need to update opname when we call def_op().
def init(
    opname,
    opmap,
    hasname,
    hasjrel,
    hasjabs,
    hasconst,
    hasarg,
    cache_format,
    specializations,
    inline_cache_entries,
    interp_only=False,
):
    def def_op(name, op):
        opmap[name] = op
        opname[op] = name

    def name_op(name, op):
        def_op(name, op)
        hasname.append(op)

    def jrel_op(name, op):
        def_op(name, op)
        hasjrel.append(op)

    def jabs_op(name, op):
        def_op(name, op)
        hasjabs.append(op)

    def_op("INVOKE_METHOD", 185)
    hasconst.append(185)
    hasarg.append(185)
    def_op("LOAD_FIELD", 186)
    hasconst.append(186)
    hasarg.append(186)
    cache_format["LOAD_FIELD"] = "{'cache': 2}"
    inline_cache_entries[186] = 2
    def_op("LOAD_OBJ_FIELD", 187)
    hasconst.append(187)
    hasarg.append(187)
    if "LOAD_FIELD" not in specializations:
        specializations["LOAD_FIELD"] = []
    specializations["LOAD_FIELD"].append("LOAD_OBJ_FIELD")
    inline_cache_entries[187] = 2
    def_op("LOAD_PRIMITIVE_FIELD", 188)
    hasconst.append(188)
    hasarg.append(188)
    if "LOAD_FIELD" not in specializations:
        specializations["LOAD_FIELD"] = []
    specializations["LOAD_FIELD"].append("LOAD_PRIMITIVE_FIELD")
    inline_cache_entries[188] = 2
    def_op("STORE_FIELD", 189)
    hasconst.append(189)
    hasarg.append(189)
    cache_format["STORE_FIELD"] = "{'cache': 2}"
    inline_cache_entries[189] = 2
    def_op("STORE_OBJ_FIELD", 190)
    hasconst.append(190)
    hasarg.append(190)
    if "STORE_FIELD" not in specializations:
        specializations["STORE_FIELD"] = []
    specializations["STORE_FIELD"].append("STORE_OBJ_FIELD")
    inline_cache_entries[190] = 2
    def_op("STORE_PRIMITIVE_FIELD", 191)
    hasconst.append(191)
    hasarg.append(191)
    if "STORE_FIELD" not in specializations:
        specializations["STORE_FIELD"] = []
    specializations["STORE_FIELD"].append("STORE_PRIMITIVE_FIELD")
    inline_cache_entries[191] = 2
    def_op("BUILD_CHECKED_LIST", 192)
    hasconst.append(192)
    hasarg.append(192)
    cache_format["BUILD_CHECKED_LIST"] = "{'cache': 2}"
    inline_cache_entries[192] = 2
    def_op("BUILD_CHECKED_LIST_CACHED", 193)
    hasconst.append(193)
    hasarg.append(193)
    if "BUILD_CHECKED_LIST" not in specializations:
        specializations["BUILD_CHECKED_LIST"] = []
    specializations["BUILD_CHECKED_LIST"].append("BUILD_CHECKED_LIST_CACHED")
    inline_cache_entries[193] = 2
    def_op("LOAD_TYPE", 194)
    hasconst.append(194)
    hasarg.append(194)
    def_op("CAST", 195)
    hasconst.append(195)
    hasarg.append(195)
    cache_format["CAST"] = "{'cache': 2}"
    inline_cache_entries[195] = 2
    def_op("CAST_CACHED", 196)
    hasconst.append(196)
    hasarg.append(196)
    if "CAST" not in specializations:
        specializations["CAST"] = []
    specializations["CAST"].append("CAST_CACHED")
    inline_cache_entries[196] = 2
    def_op("LOAD_LOCAL", 197)
    hasconst.append(197)
    hasarg.append(197)
    def_op("STORE_LOCAL", 198)
    hasconst.append(198)
    hasarg.append(198)
    cache_format["STORE_LOCAL"] = "{'cache': 1}"
    inline_cache_entries[198] = 1
    def_op("STORE_LOCAL_CACHED", 199)
    hasconst.append(199)
    hasarg.append(199)
    if "STORE_LOCAL" not in specializations:
        specializations["STORE_LOCAL"] = []
    specializations["STORE_LOCAL"].append("STORE_LOCAL_CACHED")
    inline_cache_entries[199] = 1
    def_op("PRIMITIVE_BOX", 200)
    hasarg.append(200)
    jrel_op("POP_JUMP_IF_ZERO", 201)
    hasarg.append(201)
    jrel_op("POP_JUMP_IF_NONZERO", 202)
    hasarg.append(202)
    def_op("PRIMITIVE_UNBOX", 203)
    hasarg.append(203)
    def_op("PRIMITIVE_BINARY_OP", 204)
    hasarg.append(204)
    def_op("PRIMITIVE_UNARY_OP", 205)
    hasarg.append(205)
    def_op("PRIMITIVE_COMPARE_OP", 206)
    hasarg.append(206)
    def_op("LOAD_ITERABLE_ARG", 207)
    hasarg.append(207)
    def_op("LOAD_MAPPING_ARG", 208)
    hasarg.append(208)
    def_op("INVOKE_FUNCTION", 209)
    hasconst.append(209)
    hasarg.append(209)
    cache_format["INVOKE_FUNCTION"] = "{'cache': 4}"
    inline_cache_entries[209] = 4
    def_op("INVOKE_FUNCTION_CACHED", 210)
    hasconst.append(210)
    hasarg.append(210)
    if "INVOKE_FUNCTION" not in specializations:
        specializations["INVOKE_FUNCTION"] = []
    specializations["INVOKE_FUNCTION"].append("INVOKE_FUNCTION_CACHED")
    inline_cache_entries[210] = 4
    def_op("INVOKE_INDIRECT_CACHED", 211)
    hasconst.append(211)
    hasarg.append(211)
    if "INVOKE_FUNCTION" not in specializations:
        specializations["INVOKE_FUNCTION"] = []
    specializations["INVOKE_FUNCTION"].append("INVOKE_INDIRECT_CACHED")
    inline_cache_entries[211] = 4
    if not interp_only:
        jabs_op("JUMP_IF_ZERO_OR_POP", 212)
    if not interp_only:
        jabs_op("JUMP_IF_NONZERO_OR_POP", 213)
    def_op("FAST_LEN", 214)
    hasarg.append(214)
    def_op("CONVERT_PRIMITIVE", 215)
    hasarg.append(215)
    def_op("INVOKE_NATIVE", 216)
    hasconst.append(216)
    hasarg.append(216)
    def_op("LOAD_CLASS", 217)
    hasconst.append(217)
    hasarg.append(217)
    def_op("BUILD_CHECKED_MAP", 218)
    hasconst.append(218)
    hasarg.append(218)
    cache_format["BUILD_CHECKED_MAP"] = "{'cache': 2}"
    inline_cache_entries[218] = 2
    def_op("BUILD_CHECKED_MAP_CACHED", 219)
    hasconst.append(219)
    hasarg.append(219)
    if "BUILD_CHECKED_MAP" not in specializations:
        specializations["BUILD_CHECKED_MAP"] = []
    specializations["BUILD_CHECKED_MAP"].append("BUILD_CHECKED_MAP_CACHED")
    inline_cache_entries[219] = 2
    def_op("SEQUENCE_GET", 220)
    hasarg.append(220)
    def_op("SEQUENCE_SET", 221)
    hasarg.append(221)
    def_op("LIST_DEL", 222)
    def_op("REFINE_TYPE", 223)
    hasconst.append(223)
    hasarg.append(223)
    def_op("PRIMITIVE_LOAD_CONST", 224)
    hasconst.append(224)
    hasarg.append(224)
    def_op("RETURN_PRIMITIVE", 225)
    hasarg.append(225)
    def_op("TP_ALLOC", 226)
    hasconst.append(226)
    hasarg.append(226)
    cache_format["TP_ALLOC"] = "{'cache': 2}"
    inline_cache_entries[226] = 2
    def_op("TP_ALLOC_CACHED", 227)
    hasconst.append(227)
    hasarg.append(227)
    if "TP_ALLOC" not in specializations:
        specializations["TP_ALLOC"] = []
    specializations["TP_ALLOC"].append("TP_ALLOC_CACHED")
    inline_cache_entries[227] = 2
    def_op("LOAD_METHOD_STATIC", 228)
    hasconst.append(228)
    hasarg.append(228)
    cache_format["LOAD_METHOD_STATIC"] = "{'cache': 2}"
    inline_cache_entries[228] = 2
    def_op("LOAD_METHOD_STATIC_CACHED", 230)
    hasconst.append(230)
    hasarg.append(230)
    if "LOAD_METHOD_STATIC" not in specializations:
        specializations["LOAD_METHOD_STATIC"] = []
    specializations["LOAD_METHOD_STATIC"].append("LOAD_METHOD_STATIC_CACHED")
    inline_cache_entries[230] = 2



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations

import ast
from ast import (
    AnnAssign,
    Assign,
    AST,
    AsyncFor,
    AsyncFunctionDef,
    AsyncWith,
    ClassDef,
    For,
    FunctionDef,
    If,
    Import,
    ImportFrom,
    Name,
    Try,
    While,
    With,
)
from typing import TYPE_CHECKING, Union

from .module_table import DeferredImport, ModuleTable
from .types import (
    AwaitableTypeRef,
    Class,
    DecoratedMethod,
    Function,
    InitSubclassFunction,
    ResolvedTypeRef,
    TypeEnvironment,
    TypeName,
    TypeRef,
    UnknownDecoratedMethod,
)
from .util import make_qualname, sys_hexversion_check
from .visitor import GenericVisitor

if TYPE_CHECKING:
    from .compiler import Compiler


class NestedScope:
    def __init__(self, parent_qualname: str | None) -> None:
        self.qualname: str = make_qualname(parent_qualname, "<nested>")

    def declare_class(self, node: AST, klass: Class) -> None:
        pass

    def declare_function(self, func: Function | DecoratedMethod) -> None:
        pass

    def declare_variable(self, node: AnnAssign, module: ModuleTable) -> None:
        pass

    def declare_variables(self, node: Assign, module: ModuleTable) -> None:
        pass

    # pyre-ignore[11]: Annotation `ast.TypeAlias` is not defined as a type
    def declare_type_alias(self, node: ast.TypeAlias) -> None:
        pass


TScopeTypes = Union[ModuleTable, Class, Function, NestedScope]


class DeclarationVisitor(GenericVisitor[None]):
    def __init__(
        self, mod_name: str, filename: str, symbols: Compiler, optimize: int
    ) -> None:
        module = symbols[mod_name] = ModuleTable(
            mod_name, filename, symbols, first_pass_done=False
        )
        super().__init__(module)
        self.scopes: list[TScopeTypes] = [self.module]
        self.optimize = optimize
        self.compiler = symbols
        self.type_env: TypeEnvironment = symbols.type_env

    def finish_bind(self) -> None:
        self.module.finish_bind()

    def parent_scope(self) -> TScopeTypes:
        return self.scopes[-1]

    def enter_scope(self, scope: TScopeTypes) -> None:
        self.scopes.append(scope)

    def exit_scope(self) -> None:
        self.scopes.pop()

    def qualify_name(self, name: str) -> str:
        return make_qualname(self.parent_scope().qualname, name)

    def make_nested_scope(self) -> NestedScope:
        return NestedScope(self.parent_scope().qualname)

    def enter_nested_scope(self) -> None:
        self.enter_scope(self.make_nested_scope())

    def visitAnnAssign(self, node: AnnAssign) -> None:
        self.parent_scope().declare_variable(node, self.module)

    def visitAssign(self, node: Assign) -> None:
        self.parent_scope().declare_variables(node, self.module)

    # pyre-ignore[11]: Annotation `ast.TypeAlias` is not defined as a type
    def visitTypeAlias(self, node: ast.TypeAlias) -> None:
        self.parent_scope().declare_type_alias(node)

    def visitClassDef(self, node: ClassDef) -> None:
        parent_scope = self.parent_scope()
        qualname = make_qualname(parent_scope.qualname, node.name)

        bases = [
            self.module.resolve_type(base, qualname) or self.type_env.dynamic
            for base in node.bases
        ]
        if not bases:
            bases.append(self.type_env.object)

        with self.compiler.error_sink.error_context(self.filename, node):
            klasses = []
            for base in bases:
                klasses.append(
                    base.make_subclass(
                        TypeName(self.module_name, qualname),
                        bases,
                    )
                )
            for cur_type in klasses:
                if type(cur_type) is not type(klasses[0]):
                    self.syntax_error("Incompatible subtypes", node)
            klass = klasses[0]

        for base in bases:
            if base is self.type_env.named_tuple:
                # In named tuples, the fields are actually elements
                # of the tuple, so we can't do any advanced binding against it.
                klass = self.type_env.dynamic
                break

            if base is self.type_env.protocol:
                # Protocols aren't guaranteed to exist in the actual MRO, so let's treat
                # them as dynamic to force dynamic dispatch.
                klass = self.type_env.dynamic
                break

            if base is self.type_env.typed_dict:
                # TASK(T121706684) Supporting typed dicts is tricky, similar
                # to protocols and named tuples
                klass = self.type_env.dynamic
                break

            if base.is_final:
                self.syntax_error(
                    f"Class `{klass.instance.name}` cannot subclass a Final class: `{base.instance.name}`",
                    node,
                )

        # we can't statically load classes nested inside functions
        if not isinstance(parent_scope, (ModuleTable, Class)):
            klass = self.type_env.dynamic

        self.enter_scope(
            self.make_nested_scope() if klass is self.type_env.dynamic else klass
        )
        for item in node.body:
            with self.compiler.error_sink.error_context(self.filename, item):
                self.visit(item)
        self.exit_scope()

        for d in reversed(node.decorator_list):
            if klass is self.type_env.dynamic:
                break
            with self.compiler.error_sink.error_context(self.filename, d):
                decorator = self.module.resolve_decorator(d) or self.type_env.dynamic
                klass = decorator.resolve_decorate_class(klass, d, self)

        parent_scope.declare_class(node, klass.exact_type())
        # We want the name corresponding to `C` to be the exact type when imported.
        self.module.types[node] = klass.exact_type()

    def _visitFunc(self, node: FunctionDef | AsyncFunctionDef) -> None:
        function = self._make_function(node)
        self.parent_scope().declare_function(function)

    def _make_function(self, node: FunctionDef | AsyncFunctionDef) -> Function:
        qualname = self.qualify_name(node.name)
        if node.name == "__init_subclass__":
            func = InitSubclassFunction(
                node, self.module, self.return_type_ref(node, qualname)
            )
            parent_scope = self.parent_scope()
            if isinstance(parent_scope, Class):
                parent_scope.has_init_subclass = True
        else:
            func = Function(node, self.module, self.return_type_ref(node, qualname))
        self.enter_scope(func)
        for item in node.body:
            self.visit(item)
        self.exit_scope()

        func_type = func
        if node.decorator_list:
            # Since we haven't resolved decorators yet (until finish_bind), we
            # don't know what type we should ultimately set for this node;
            # Function.finish_bind() will likely override this.
            func_type = UnknownDecoratedMethod(func, node.decorator_list[0])

        self.module.types[node] = func_type
        return func

    def visitFunctionDef(self, node: FunctionDef) -> None:
        self._visitFunc(node)

    def visitAsyncFunctionDef(self, node: AsyncFunctionDef) -> None:
        self._visitFunc(node)

    def return_type_ref(
        self, node: FunctionDef | AsyncFunctionDef, requester: str
    ) -> TypeRef:
        ann = node.returns
        if not ann:
            res = ResolvedTypeRef(self.type_env.dynamic)
        else:
            res = TypeRef(self.module, requester, ann)
        if isinstance(node, AsyncFunctionDef):
            res = AwaitableTypeRef(res, self.module.compiler)
        return res

    def visitImport(self, node: Import) -> None:
        for name in node.names:
            asname = name.asname
            if asname is None:
                top_level_module = name.name.split(".")[0]
                self.module.declare_import(
                    top_level_module,
                    None,
                    DeferredImport(
                        self.module,
                        name.name,
                        top_level_module,
                        self.optimize,
                        self.compiler,
                        mod_to_return=top_level_module,
                    ),
                )
            else:
                self.module.declare_import(
                    asname,
                    None,
                    DeferredImport(
                        self.module, name.name, asname, self.optimize, self.compiler
                    ),
                )

    def visitImportFrom(self, node: ImportFrom) -> None:
        mod_name = node.module
        if not mod_name or node.level:
            raise NotImplementedError("relative imports aren't supported")
        for name in node.names:
            child_name = name.asname or name.name
            self.module.declare_import(
                child_name,
                (mod_name, name.name),
                DeferredImport(
                    self.module,
                    mod_name,
                    name.name,
                    self.optimize,
                    self.compiler,
                    name.name,
                ),
            )

    # We don't pick up declarations in nested statements
    def visitFor(self, node: For) -> None:
        self.enter_nested_scope()
        self.generic_visit(node)
        self.exit_scope()

    def visitAsyncFor(self, node: AsyncFor) -> None:
        self.enter_nested_scope()
        self.generic_visit(node)
        self.exit_scope()

    def visitWhile(self, node: While) -> None:
        self.enter_nested_scope()
        self.generic_visit(node)
        self.exit_scope()

    def visitIf(self, node: If) -> None:
        test = node.test
        if isinstance(test, Name) and test.id == "TYPE_CHECKING":
            self.visit_list(node.body)
        else:
            result = sys_hexversion_check(node)
            # We should check the version if provided
            if result is not None:
                self.module.mark_known_boolean_test(test, value=bool(result))
                if result:
                    self.visit_list(node.body)
                else:
                    self.visit_list(node.orelse)
                return
            else:
                self.enter_nested_scope()
                self.visit_list(node.body)
                self.exit_scope()

        if node.orelse:
            self.enter_nested_scope()
            self.visit_list(node.orelse)
            self.exit_scope()

    def visitWith(self, node: With) -> None:
        self.enter_nested_scope()
        self.generic_visit(node)
        self.exit_scope()

    def visitAsyncWith(self, node: AsyncWith) -> None:
        self.enter_nested_scope()
        self.generic_visit(node)
        self.exit_scope()

    def visitTry(self, node: Try) -> None:
        self.enter_nested_scope()
        self.generic_visit(node)
        self.exit_scope()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict

from __future__ import annotations

from ast import AST
from contextlib import contextmanager, nullcontext
from typing import ContextManager, Generator, Generic, Sequence, TYPE_CHECKING, TypeVar

from ..visitor import ASTVisitor

if TYPE_CHECKING:
    from ..errors import ErrorSink
    from .compiler import Compiler
    from .module_table import ModuleTable
    from .types import TypeEnvironment


TVisitRet = TypeVar("TVisitRet", covariant=True)


class GenericVisitor(ASTVisitor, Generic[TVisitRet]):
    def __init__(self, module: ModuleTable) -> None:
        super().__init__()
        self.module = module
        self.module_name: str = module.name
        self.filename: str = module.filename
        self.compiler: Compiler = module.compiler
        self.error_sink: ErrorSink = module.compiler.error_sink
        self.type_env: TypeEnvironment = module.compiler.type_env
        # the qualname that should be the "requester" of types used (for dep tracking)
        self._context_qualname: str = ""
        # if true, all deps tracked in visiting should be considered decl deps
        self.force_decl_deps: bool = False

    @property
    def context_qualname(self) -> str:
        return self._context_qualname

    @contextmanager
    def temporary_context_qualname(
        self, qualname: str | None, force_decl: bool = False
    ) -> Generator[None, None, None]:
        old_qualname = self._context_qualname
        self._context_qualname = qualname or ""
        old_decl = self.force_decl_deps
        self.force_decl_deps = force_decl
        try:
            yield
        finally:
            self._context_qualname = old_qualname
            self.force_decl_deps = old_decl

    def record_dependency(self, source: tuple[str, str]) -> None:
        self.module.record_dependency(
            self.context_qualname, source, force_decl=self.force_decl_deps
        )

    def visit(self, node: AST, *args: object) -> TVisitRet:
        # if we have a sequence of nodes, don't catch TypedSyntaxError here;
        # walk_list will call us back with each individual node in turn and we
        # can catch errors and add node info then.
        ctx = self.error_context(node) if isinstance(node, AST) else nullcontext()
        with ctx:
            if not args:
                # pyre-ignore: ASTVisitor is not generic yet, can't assert the result is
                # TVisitRet.
                return super().visit(node)
            # pyre-ignore: ASTVisitor is not generic yet, can't assert the result is
            # TVisitRet.
            return super().visit(node, *args)

    def syntax_error(self, msg: str, node: AST) -> None:
        return self.error_sink.syntax_error(msg, self.filename, node)

    def perf_warning(self, msg: str, node: AST) -> None:
        return self.error_sink.perf_warning(msg, self.filename, node)

    def error_context(self, node: AST | None) -> ContextManager[None]:
        if node is None:
            return nullcontext()
        return self.error_sink.error_context(self.filename, node)

    @contextmanager
    def temporary_error_sink(self, sink: ErrorSink) -> Generator[None, None, None]:
        orig_sink = self.error_sink
        self.error_sink = sink
        try:
            yield
        finally:
            self.error_sink = orig_sink



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations

import ast
from typing import final, TypeVar


FunctionDefNode = TypeVar("FunctionDefNode", ast.FunctionDef, ast.AsyncFunctionDef)


def remove_annotations(node: ast.AST) -> ast.Module:
    return ast.fix_missing_locations(AnnotationRemover().visit(node))


def _copy_attrs(src: ast.AST, dest: ast.AST) -> None:
    """
    Copies line and column info from one node to another.
    """
    # pyre-fixme[16]: `AST` has no attribute `lineno`.
    dest.lineno = src.lineno
    # pyre-fixme[16]: `AST` has no attribute `end_lineno`.
    dest.end_lineno = src.end_lineno
    # pyre-fixme[16]: `AST` has no attribute `col_offset`.
    dest.col_offset = src.col_offset
    # pyre-fixme[16]: `AST` has no attribute `end_col_offset`.
    dest.end_col_offset = src.end_col_offset


@final
class AnnotationRemover(ast.NodeTransformer):
    def visit_single_arg(self, arg: ast.arg) -> ast.arg:
        arg.annotation = None
        return arg

    def visit_fn_arguments(self, node: ast.arguments) -> ast.arguments:
        if node.posonlyargs:
            node.posonlyargs = [self.visit_single_arg(a) for a in node.posonlyargs]
        if node.args:
            node.args = [self.visit_single_arg(a) for a in node.args]
        if node.kwonlyargs:
            node.kwonlyargs = [self.visit_single_arg(a) for a in node.kwonlyargs]
        vararg = node.vararg
        if vararg:
            node.vararg = self.visit_single_arg(vararg)
        kwarg = node.kwarg
        if kwarg:
            node.kwarg = self.visit_single_arg(kwarg)
        return node

    def visit_function(self, node: FunctionDefNode) -> FunctionDefNode:
        node.arguments = self.visit_fn_arguments(node.args)
        node.returns = None
        node.decorator_list = [
            self.visit(decorator) for decorator in node.decorator_list
        ]
        return node

    def visit_FunctionDef(self, node: FunctionDefNode) -> FunctionDefNode:
        return self.visit_function(node)

    def visit_AsyncFunctionDef(self, node: FunctionDefNode) -> FunctionDefNode:
        return self.visit_function(node)

    def visit_AnnAssign(self, node: ast.AnnAssign) -> ast.Assign:
        # Here, we replace `x: A = a` with just `x = a`. In case there's no value,
        # we set the value to an `...`. E.g: `x: A` changes to `x = ...`.
        #
        # We need to be a little careful about ensuring that newly created nodes
        # get the line and column information copied to them. This helps us avoid
        # an extra pass over the AST with ast.fix_missing_locations()
        value = node.value
        if value is None:
            # pyre-fixme[20]: Argument `value` expected.
            value = ast.Constant(...)
            value.kind = None
            _copy_attrs(node, value)

        assign = ast.Assign(targets=[node.target], value=value, type_comment=None)

        _copy_attrs(node, assign)
        return assign



# Copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict

from .remove_annotations import AnnotationRemover, remove_annotations  # noqa: F401
from .rewriter import rewrite, StrictModuleRewriter  # noqa: F401



# Portions copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations

import ast
import operator
from ast import cmpop, Constant, copy_location
from dataclasses import dataclass
from typing import Any, Callable, Iterable, Mapping

from .visitor import ASTRewriter


class PyLimits:
    MAX_INT_SIZE = 128
    MAX_COLLECTION_SIZE = 256
    MAX_STR_SIZE = 4096
    MAX_TOTAL_ITEMS = 1024


# pyre-fixme[5]: Global annotation cannot contain `Any`.
UNARY_OPS: Mapping[type[ast.unaryop], Callable[[Any], object]] = {
    ast.Invert: operator.invert,
    ast.Not: operator.not_,
    ast.UAdd: operator.pos,
    ast.USub: operator.neg,
}
INVERSE_OPS: Mapping[type[cmpop], type[cmpop]] = {
    ast.Is: ast.IsNot,
    ast.IsNot: ast.Is,
    ast.In: ast.NotIn,
    ast.NotIn: ast.In,
}

BIN_OPS: Mapping[type[ast.operator], Callable[[object, object], object]] = {
    ast.Add: operator.add,
    ast.Sub: operator.sub,
    ast.Mult: lambda lhs, rhs: safe_multiply(lhs, rhs, PyLimits),
    ast.Div: operator.truediv,
    ast.FloorDiv: operator.floordiv,
    ast.Mod: lambda lhs, rhs: safe_mod(lhs, rhs, PyLimits),
    ast.Pow: lambda lhs, rhs: safe_power(lhs, rhs, PyLimits),
    ast.LShift: lambda lhs, rhs: safe_lshift(lhs, rhs, PyLimits),
    ast.RShift: operator.rshift,
    ast.BitOr: operator.or_,
    ast.BitXor: operator.xor,
    ast.BitAnd: operator.and_,
}


class DefaultLimits:
    MAX_INT_SIZE = 128
    MAX_COLLECTION_SIZE = 20
    MAX_STR_SIZE = 20
    MAX_TOTAL_ITEMS = 1024


LimitsType = type[PyLimits] | type[DefaultLimits]


# pyre-fixme[2]: Parameter annotation cannot be `Any`.
def safe_lshift(left: Any, right: Any, limits: LimitsType = DefaultLimits) -> object:
    if isinstance(left, int) and isinstance(right, int) and left and right:
        lbits = left.bit_length()
        if (
            right < 0
            or right > limits.MAX_INT_SIZE
            or lbits > limits.MAX_INT_SIZE - right
        ):
            raise OverflowError()

    return left << right


def check_complexity(obj: object, limit: int) -> int:
    if isinstance(obj, (frozenset, tuple)):
        limit -= len(obj)
        for item in obj:
            limit = check_complexity(item, limit)
            if limit < 0:
                break

    return limit


# pyre-fixme[2]: Parameter annotation cannot be `Any`.
def safe_multiply(left: Any, right: Any, limits: LimitsType = DefaultLimits) -> object:
    if isinstance(left, int) and isinstance(right, int) and left and right:
        lbits = left.bit_length()
        rbits = right.bit_length()
        if lbits + rbits > limits.MAX_INT_SIZE:
            raise OverflowError()
    elif isinstance(left, int) and isinstance(right, (tuple, frozenset)):
        rsize = len(right)
        if rsize:
            if left < 0 or left > limits.MAX_COLLECTION_SIZE / rsize:
                raise OverflowError()
            if left:
                if check_complexity(right, limits.MAX_TOTAL_ITEMS // left) < 0:
                    raise OverflowError()
    elif isinstance(left, int) and isinstance(right, (str, bytes)):
        rsize = len(right)
        if rsize:
            if left < 0 or left > limits.MAX_STR_SIZE / rsize:
                raise OverflowError()
    elif isinstance(right, int) and isinstance(left, (tuple, frozenset, str, bytes)):
        return safe_multiply(right, left, limits)

    return left * right


# pyre-fixme[2]: Parameter annotation cannot be `Any`.
def safe_power(left: Any, right: Any, limits: LimitsType = DefaultLimits) -> object:
    if isinstance(left, int) and isinstance(right, int) and left and right > 0:
        lbits = left.bit_length()
        if lbits > limits.MAX_INT_SIZE / right:
            raise OverflowError()

    return left**right


# pyre-fixme[2]: Parameter annotation cannot be `Any`.
def safe_mod(left: Any, right: Any, limits: LimitsType = DefaultLimits) -> object:
    if isinstance(left, (str, bytes)):
        raise OverflowError()

    return left % right


class AstOptimizer(ASTRewriter):
    def __init__(self, optimize: bool = False, string_anns: bool = False) -> None:
        super().__init__()
        self.optimize = optimize
        self.string_anns = string_anns

    def skip_field(self, node: ast.AST, field: str) -> bool:
        if self.string_anns:
            if (
                isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef))
                and field == "returns"
            ):
                return True
            if isinstance(node, ast.arg) and field == "annotation":
                return True
            if isinstance(node, ast.AnnAssign) and field == "annotation":
                return True
        return False

    def visitUnaryOp(self, node: ast.UnaryOp) -> ast.expr:
        op = self.visit(node.operand)
        if isinstance(op, Constant):
            conv = UNARY_OPS[type(node.op)]
            try:
                return copy_location(Constant(conv(op.value)), node)
            except Exception:
                pass
        elif (
            isinstance(node.op, ast.Not)
            and isinstance(op, ast.Compare)
            and len(op.ops) == 1
        ):
            cmp_op = op.ops[0]
            new_op = INVERSE_OPS.get(type(cmp_op))
            if new_op is not None:
                return self.update_node(op, ops=[new_op()])

        return self.update_node(node, operand=op)

    def visitBinOp(self, node: ast.BinOp) -> ast.expr:
        left = self.visit(node.left)
        right = self.visit(node.right)

        if isinstance(left, Constant) and isinstance(right, Constant):
            handler = BIN_OPS.get(type(node.op))
            if handler is not None:
                try:
                    return copy_location(
                        Constant(handler(left.value, right.value)), node
                    )
                except Exception:
                    pass

        return self.update_node(node, left=left, right=right)

    def makeConstTuple(self, elts: Iterable[ast.expr]) -> Constant | None:
        if all(isinstance(elt, Constant) for elt in elts):
            # pyre-ignore[16]: each elt is a constant at this point.
            return Constant(tuple(elt.value for elt in elts))

        return None

    def visitTuple(self, node: ast.Tuple) -> ast.expr:
        elts = self.walk_list(node.elts)

        if isinstance(node.ctx, ast.Load):
            # pyre-ignore[6]: Can't type walk_list fully yet.
            res = self.makeConstTuple(elts)
            if res is not None:
                return copy_location(res, node)

        return self.update_node(node, elts=elts)

    def visitSubscript(self, node: ast.Subscript) -> ast.expr:
        value = self.visit(node.value)
        slice = self.visit(node.slice)

        if (
            isinstance(node.ctx, ast.Load)
            and isinstance(value, Constant)
            and isinstance(slice, Constant)
        ):
            try:
                return copy_location(
                    Constant(value.value[slice.value]),
                    node,
                )
            except Exception:
                pass

        return self.update_node(node, value=value, slice=slice)

    def _visitIter(self, node: ast.expr) -> ast.expr:
        if isinstance(node, ast.List):
            elts = self.walk_list(node.elts)
            # pyre-ignore[6]: Can't type walk_list fully yet.
            res = self.makeConstTuple(elts)
            if res is not None:
                return copy_location(res, node)
            if not any(isinstance(e, ast.Starred) for e in elts):
                # pyre-fixme[6]: For 1st argument expected `List[expr]` but got
                #  `Sequence[expr]`.
                return copy_location(ast.Tuple(elts=elts, ctx=node.ctx), node)
            return self.update_node(node, elts=elts)
        elif isinstance(node, ast.Set):
            elts = self.walk_list(node.elts)
            # pyre-ignore[6]: Can't type walk_list fully yet.
            res = self.makeConstTuple(elts)
            if res is not None:
                return copy_location(Constant(frozenset(res.value)), node)

            return self.update_node(node, elts=elts)

        return self.generic_visit(node)

    def visitcomprehension(self, node: ast.comprehension) -> ast.comprehension:
        target = self.visit(node.target)
        iter = self.visit(node.iter)
        assert isinstance(iter, ast.expr)
        ifs = self.walk_list(node.ifs)
        iter = self._visitIter(iter)

        return self.update_node(node, target=target, iter=iter, ifs=ifs)

    def visitFor(self, node: ast.For) -> ast.For:
        target = self.visit(node.target)
        iter = self.visit(node.iter)
        assert isinstance(iter, ast.expr)
        body = self.walk_list(node.body)
        orelse = self.walk_list(node.orelse)

        iter = self._visitIter(iter)
        return self.update_node(
            node, target=target, iter=iter, body=body, orelse=orelse
        )

    def visitCompare(self, node: ast.Compare) -> ast.expr:
        left = self.visit(node.left)
        comparators = self.walk_list(node.comparators)

        if isinstance(node.ops[-1], (ast.In, ast.NotIn)):
            # pyre-ignore[6]: Can't type walk_list fully yet.
            new_iter = self._visitIter(comparators[-1])
            if new_iter is not None and new_iter is not comparators[-1]:
                comparators = list(comparators)
                comparators[-1] = new_iter

        return self.update_node(node, left=left, comparators=comparators)

    def visitName(self, node: ast.Name) -> ast.Name | ast.Constant:
        if node.id == "__debug__":
            return copy_location(Constant(not self.optimize), node)

        return self.generic_visit(node)

    def visitNamedExpr(self, node: ast.NamedExpr) -> ast.NamedExpr:
        return self.generic_visit(node)


F_LJUST: int = 1 << 0
F_SIGN: int = 1 << 1
F_BLANK: int = 1 << 2
F_ALT: int = 1 << 3
F_ZERO: int = 1 << 4

FLAG_DICT: dict[str, int] = {
    "-": F_LJUST,
    "+": F_SIGN,
    " ": F_BLANK,
    "#": F_ALT,
    "0": F_ZERO,
}
MAXDIGITS = 3


class UnsupportedFormat(Exception):
    pass


@dataclass(frozen=True)
class FormatInfo:
    spec: str
    flags: int = 0
    width: int | None = None
    prec: int | None = None

    def as_formatted_value(self, arg: ast.expr) -> ast.FormattedValue | None:
        if self.spec not in "sra":
            return None

        res = []
        if not (self.flags & F_LJUST) and self.width:
            res.append(">")
        if self.width is not None:
            res.append(str(self.width))
        if self.prec is not None:
            res.append(f".{self.prec}")

        return copy_location(
            ast.FormattedValue(
                arg,
                ord(self.spec),
                copy_location(ast.Constant("".join(res)), arg) if res else None,
            ),
            arg,
        )


class FormatParser:
    def __init__(self, val: str) -> None:
        self.val = val
        self.size: int = len(val)
        self.pos = 0

    def next_ch(self) -> str:
        """Gets the next character in the format string, or raises an
        exception if we've run out of characters"""
        if self.pos >= self.size:
            raise UnsupportedFormat()
        self.pos += 1
        return self.val[self.pos]

    def parse_int(self) -> int:
        """Parses an integer from the format string"""
        res = 0
        digits = 0
        ch = self.val[self.pos]
        while "0" <= ch <= "9":
            res = res * 10 + ord(ch) - ord("0")
            ch = self.next_ch()
            digits += 1
            if digits >= MAXDIGITS:
                raise UnsupportedFormat()
        return res

    def parse_flags(self, ch: str) -> int:
        """Parse any flags (-, +, " ", #, 0)"""
        flags = 0
        while True:
            flag_val = FLAG_DICT.get(ch)
            if flag_val is not None:
                flags |= flag_val
                ch = self.next_ch()
            else:
                break
        return flags

    def parse_str(self) -> str:
        """Parses a string component of the format string up to a %"""
        has_percents = False
        start = self.pos
        while self.pos < self.size:
            ch = self.val[self.pos]
            if ch != "%":
                self.pos += 1
            elif self.pos + 1 < self.size and self.val[self.pos + 1] == "%":
                has_percents = True
                self.pos += 2
            else:
                break

        component = self.val[start : self.pos]
        if has_percents:
            component = component.replace("%%", "%")
        return component

    def enum_components(self) -> Iterable[str | FormatInfo]:
        """Enumerates the components of the format string and returns a stream
        of interleaved strings and FormatInfo objects"""
        # Parse the string up to the format specifier
        ch = None
        while self.pos < self.size:
            yield self.parse_str()

            if self.pos == self.size:
                return

            assert self.val[self.pos] == "%"

            flags = self.parse_flags(self.next_ch())

            # Parse width
            width = None
            if "0" <= self.val[self.pos] <= "9":
                width = self.parse_int()

            prec = None
            if self.val[self.pos] == ".":
                self.next_ch()
                prec = self.parse_int()

            yield FormatInfo(self.val[self.pos], flags, width, prec)
            self.pos += 1


def enum_format_str_components(val: str) -> Iterable[str | FormatInfo]:
    return FormatParser(val).enum_components()


def set_no_locations(node: ast.expr) -> ast.expr:
    node.lineno = -1
    node.end_lineno = -1
    node.col_offset = -1
    node.end_col_offset = -1
    return node


class AstOptimizer312(AstOptimizer):
    def visitBinOp(self, node: ast.BinOp) -> ast.expr:
        res = super().visitBinOp(node)

        if (
            not isinstance(res, ast.BinOp)
            or not isinstance(res.op, ast.Mod)
            or not isinstance(res.left, ast.Constant)
            or not isinstance(res.left.value, str)
            or not isinstance(res.right, ast.Tuple)
            or any(isinstance(e, ast.Starred) for e in res.right.elts)
        ):
            return res

        return self.optimize_format(res)

    def optimize_format(self, node: ast.BinOp) -> ast.expr:
        left = node.left
        right = node.right
        assert isinstance(left, ast.Constant)
        assert isinstance(right, ast.Tuple)
        assert isinstance(left.value, str)

        try:
            seq = []
            cnt = 0
            for item in enum_format_str_components(left.value):
                if isinstance(item, str):
                    if item:
                        seq.append(set_no_locations(ast.Constant(item)))
                    continue

                if cnt >= len(right.elts):
                    # More format units than items.
                    return node

                formatted = item.as_formatted_value(right.elts[cnt])
                if formatted is None:
                    return node
                seq.append(formatted)
                cnt += 1

            if cnt < len(right.elts):
                # More items than format units.
                return node

            return copy_location(ast.JoinedStr(seq), node)

        except UnsupportedFormat:
            return node


class AstOptimizer314(AstOptimizer312):
    def has_starred(self, e: ast.Tuple) -> bool:
        return any(isinstance(e, ast.Starred) for e in e.elts)

    def visitUnaryOp(self, node: ast.UnaryOp) -> ast.expr:
        op = self.visit(node.operand)
        return self.update_node(node, operand=op)

    def visitBinOp(self, node: ast.BinOp) -> ast.expr:
        lhs = self.visit(node.left)
        rhs = self.visit(node.right)
        if (
            isinstance(lhs, ast.Constant)
            and isinstance(rhs, ast.Tuple)
            and isinstance(lhs.value, str)
            and not self.has_starred(rhs)
        ):
            return self.optimize_format(self.update_node(node, left=lhs, right=rhs))

        return self.update_node(node, left=lhs, right=rhs)

    def visitSubscript(self, node: ast.Subscript) -> ast.expr:
        value = self.visit(node.value)
        slice = self.visit(node.slice)

        return self.update_node(node, value=value, slice=slice)

    def visitMatchValue(self, node: ast.MatchValue) -> ast.MatchValue:
        return self.update_node(node, value=self.fold_const_match_patterns(node.value))

    def visitMatchMapping(self, node: ast.MatchMapping) -> ast.MatchMapping:
        keys = [self.fold_const_match_patterns(key) for key in node.keys]
        patterns = self.walk_list(node.patterns)
        return self.update_node(node, keys=keys, patterns=patterns)

    def fold_const_match_patterns(self, node: ast.expr) -> ast.expr:
        if isinstance(node, ast.UnaryOp):
            if isinstance(node.op, ast.USub) and isinstance(node.operand, ast.Constant):
                return super().visitUnaryOp(node)
        elif isinstance(node, ast.BinOp) and isinstance(node.op, (ast.Add, ast.Sub)):
            if isinstance(node.right, ast.Constant):
                node = self.update_node(
                    node, left=self.fold_const_match_patterns(node.left)
                )
                if isinstance(node.left, ast.Constant):
                    return super().visitBinOp(node)

        return node

    def visitTuple(self, node: ast.Tuple) -> ast.expr:
        elts = self.walk_list(node.elts)

        return self.update_node(node, elts=elts)

    def _visitIter(self, node: ast.expr) -> ast.expr:
        # This optimization has been removed in 3.14
        return node



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations

import ast
from collections.abc import Sequence
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .types import Value

    RefinedFields = dict[str, tuple[Value, int, set[ast.AST]]]


# A refined field consists of a refined type in addition to the node that
# refined the field. The node information is used to hoist reads during codegen.


class TypeState:
    def __init__(self) -> None:
        self.local_types: dict[str, Value] = {}
        self.refined_fields: dict[str, RefinedFields] = {}

    def copy(self) -> TypeState:
        type_state = TypeState()
        type_state.local_types = dict(self.local_types)
        type_state.refined_fields = dict(self.refined_fields)
        return type_state

    def __repr__(self) -> str:
        return f"TypeState({self.local_types}, {self.refined_fields})"


class NarrowingEffect:
    """captures type narrowing effects on variables"""

    def and_(self, other: NarrowingEffect) -> NarrowingEffect:
        if other is NO_EFFECT:
            return self

        return AndEffect(self, other)

    def or_(self, other: NarrowingEffect) -> NarrowingEffect:
        if other is NO_EFFECT:
            return self

        return OrEffect(self, other)

    def not_(self) -> NarrowingEffect:
        return NegationEffect(self)

    def apply(
        self,
        type_state: TypeState,
        type_state_nodes: dict[str, ast.AST] | None = None,
    ) -> None:
        """applies the given effect in the target scope. if `type_state_nodes` is passed, populates
        it with the underlying name or attribute nodes"""
        pass

    def undo(self, type_state: TypeState) -> None:
        """restores the type to its original value"""
        pass

    def reverse(
        self,
        type_state: TypeState,
        type_state_nodes: dict[str, ast.AST] | None = None,
    ) -> None:
        """applies the reverse of the scope or reverts it if
        there is no reverse"""
        self.undo(type_state)


class AndEffect(NarrowingEffect):
    def __init__(self, *effects: NarrowingEffect) -> None:
        self.effects: Sequence[NarrowingEffect] = effects

    def and_(self, other: NarrowingEffect) -> NarrowingEffect:
        if other is NO_EFFECT:
            return self
        elif isinstance(other, AndEffect):
            return AndEffect(*self.effects, *other.effects)

        return AndEffect(*self.effects, other)

    def apply(
        self,
        type_state: TypeState,
        type_state_nodes: dict[str, ast.AST] | None = None,
    ) -> None:
        for effect in self.effects:
            effect.apply(type_state, type_state_nodes)

    def undo(self, type_state: TypeState) -> None:
        """restores the type to its original value"""
        for effect in self.effects:
            effect.undo(type_state)


class OrEffect(NarrowingEffect):
    def __init__(self, *effects: NarrowingEffect) -> None:
        self.effects: Sequence[NarrowingEffect] = effects

    def and_(self, other: NarrowingEffect) -> NarrowingEffect:
        if other is NO_EFFECT:
            return self
        elif isinstance(other, OrEffect):
            return OrEffect(*self.effects, *other.effects)

        return OrEffect(*self.effects, other)

    def reverse(
        self,
        type_state: TypeState,
        type_state_nodes: dict[str, ast.AST] | None = None,
    ) -> None:
        for effect in self.effects:
            effect.reverse(type_state, type_state_nodes)

    def undo(self, type_state: TypeState) -> None:
        """restores the type to its original value"""
        for effect in self.effects:
            effect.undo(type_state)


class NoEffect(NarrowingEffect):
    def union(self, other: NarrowingEffect) -> NarrowingEffect:
        return other


# Singleton instance for no effects
NO_EFFECT = NoEffect()


class NegationEffect(NarrowingEffect):
    def __init__(self, negated: NarrowingEffect) -> None:
        self.negated = negated

    def not_(self) -> NarrowingEffect:
        return self.negated

    def apply(
        self,
        type_state: TypeState,
        type_state_nodes: dict[str, ast.AST] | None = None,
    ) -> None:
        self.negated.reverse(type_state, type_state_nodes)

    def undo(self, type_state: TypeState) -> None:
        self.negated.undo(type_state)

    def reverse(
        self,
        type_state: TypeState,
        type_state_nodes: dict[str, ast.AST] | None = None,
    ) -> None:
        self.negated.apply(type_state, type_state_nodes)



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations

import ast
from ast import (
    AnnAssign,
    Assign,
    AST,
    AsyncFunctionDef,
    Attribute,
    ClassDef,
    Delete,
    ExceptHandler,
    FunctionDef,
    Global,
    Import,
    ImportFrom,
    Module,
    Name,
)
from symtable import SymbolTable
from typing import final, MutableMapping

from ..consts import CO_FUTURE_ANNOTATIONS
from ..pycodegen import find_futures
from .common import (
    get_symbol_map,
    imported_name,
    ScopeStack,
    StrictModuleError,
    SymbolMap,
    SymbolScope,
)
from .rewriter.rewriter import SymbolVisitor

CLASS_ATTR_CONFLICT_EXCEPTION = "ClassAttributesConflictException"


class TransformerScope:
    def visit_Assign(self, node: Assign) -> None:
        pass

    def visit_AnnAssign(self, node: AnnAssign) -> None:
        pass

    def loaded(self, name: str) -> None:
        pass

    def stored(self, name: str) -> None:
        pass


@final
class ClassScope(TransformerScope):
    def __init__(self) -> None:
        self.instance_fields: set[str] = set()
        self.class_fields: set[str] = set()

    def visit_AnnAssign(self, node: AnnAssign) -> None:
        target = node.target
        if isinstance(target, Name):
            if node.value is None:
                self.instance_fields.add(target.id)
            else:
                self.class_fields.add(target.id)

    def stored(self, name: str) -> None:
        self.class_fields.add(name)


@final
class FunctionScope(TransformerScope):
    def __init__(self, node: FunctionDef, parent: TransformerScope) -> None:
        self.node = node
        self.parent = parent

    def visit_AnnAssign(self, node: AnnAssign) -> None:
        parent = self.parent
        target = node.target
        if (
            isinstance(self.node, FunctionDef)
            and self.node.name == "__init__"
            and self.node.args.args
            and isinstance(parent, ClassScope)
            and isinstance(target, Attribute)
        ):
            self.add_attr_name(target, parent)

    def add_attr_name(self, target: Attribute, parent: ClassScope) -> None:
        """records self.name = ... when salf matches the 1st parameter"""
        value = target.value
        node = self.node
        if (
            isinstance(node, FunctionDef)
            and isinstance(value, Name)
            and value.id == node.args.args[0].arg
        ):
            parent.instance_fields.add(target.attr)

    def visit_Assign(self, node: Assign) -> None:
        parent = self.parent
        if (
            isinstance(self.node, FunctionDef)
            and self.node.name == "__init__"
            and isinstance(parent, ClassScope)
            and self.node.args.args
        ):
            for target in node.targets:
                if not isinstance(target, Attribute):
                    continue
                self.add_attr_name(target, parent)


@final
class ClassConflictChecker(SymbolVisitor[object, TransformerScope]):
    def __init__(
        self,
        symbols: SymbolTable,
        symbol_map: SymbolMap,
        filename: str,
        flags: int,
    ) -> None:
        super().__init__(
            ScopeStack(
                self.make_scope(symbols, None),
                symbol_map=symbol_map,
                scope_factory=self.make_scope,
            ),
        )
        self.filename = filename
        self.flags = flags

    @property
    def skip_annotations(self) -> bool:
        return bool(self.flags & CO_FUTURE_ANNOTATIONS)

    def error(self, names: list[str], lineno: int, col: int, filename: str) -> None:
        MSG: str = "Class member conflicts with instance member: {names}"
        raise StrictModuleError(MSG.format(names=names), filename, lineno, col)

    def make_scope(
        self,
        symtable: SymbolTable,
        node: AST | None,
        vars: MutableMapping[str, object] | None = None,
    ) -> SymbolScope[object, TransformerScope]:
        if isinstance(node, FunctionDef):
            data = FunctionScope(node, self.scopes.scopes[-1].scope_data)
        elif isinstance(node, ClassDef):
            data = ClassScope()
        else:
            data = TransformerScope()
        return SymbolScope(symtable, data)

    def visit_Name(self, node: Name) -> None:
        scope = self.scope_for(node.id).scope_data
        if isinstance(node.ctx, ast.Load):
            scope.loaded(node.id)
        else:
            scope.stored(node.id)

    def visit_ExceptHandler(self, node: ExceptHandler) -> None:
        self.generic_visit(node)
        name = node.name
        if name is not None:
            self.scope_for(name).scope_data.stored(name)

    def visit_Delete(self, node: Delete) -> None:
        for target in node.targets:
            if isinstance(target, ast.Name):
                self.scope_for(target.id).scope_data.stored(target.id)

    def visit_Global(self, node: Global) -> None:
        if self.scopes.in_class_scope:
            for name in node.names:
                if name == "__annotations__":
                    self.error(
                        ["__annotations__"], node.lineno, node.col_offset, self.filename
                    )

    def visit_ClassDef(self, node: ClassDef) -> None:
        self.visit_Class_Outer(node)
        class_scope = self.visit_Class_Inner(node).scope_data
        assert isinstance(class_scope, ClassScope)

        overlap = class_scope.instance_fields.intersection(class_scope.class_fields)
        if overlap:
            self.error(list(overlap), node.lineno, node.col_offset, self.filename)

        self.scope_for(node.name).scope_data.stored(node.name)

    def visit_FunctionDef(self, node: FunctionDef) -> None:
        self.visit_Func_Outer(node)

        func_scope = self.visit_Func_Inner(node)

        self.scopes.current[node.name] = func_scope.scope_data

        self.scope_for(node.name).scope_data.stored(node.name)

    def visit_AsyncFunctionDef(self, node: AsyncFunctionDef) -> None:
        self.visit_Func_Outer(node)

        self.visit_Func_Inner(node)

        self.scope_for(node.name).scope_data.stored(node.name)

    def visit_Import(self, node: Import) -> None:
        for name in node.names:
            self.scope_for(imported_name(name)).scope_data.stored(imported_name(name))
        return self.generic_visit(node)

    def visit_ImportFrom(self, node: ImportFrom) -> None:
        if node.level == 0 and node.module is not None:
            for name in node.names:
                self.scope_for(name.asname or name.name).scope_data.stored(
                    name.asname or name.name
                )

    def visit_Assign(self, node: Assign) -> None:
        self.scopes.scopes[-1].scope_data.visit_Assign(node)
        self.generic_visit(node)

    def visit_AnnAssign(self, node: AnnAssign) -> None:
        self.scopes.scopes[-1].scope_data.visit_AnnAssign(node)
        value = node.value
        if value is not None:
            self.visit(node.target)
            self.visit(value)
            if not self.skip_annotations:
                self.visit(node.annotation)

    def visit_arg(self, node: ast.arg) -> None:
        if not self.skip_annotations:
            self.generic_visit(node)


def check_class_conflict(
    node: Module,
    filename: str,
    symbols: SymbolTable,
) -> None:
    symbol_map = get_symbol_map(node, symbols)
    flags = find_futures(0, node)
    visitor = ClassConflictChecker(symbols, symbol_map, filename=filename, flags=flags)
    visitor.visit(node)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# pyre-unsafe
# @noautodeps

import datetime
import glob
import os
import os.path
import re
import shutil
import subprocess
import sys
import sysconfig
from enum import Enum
from functools import lru_cache
from typing import Callable

from setuptools import Extension, find_packages, setup
from setuptools.command.build import build as build
from setuptools.command.build_ext import build_ext
from setuptools.command.build_py import build_py
from setuptools.dist import Distribution

CHECKOUT_ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
SOURCE_DIR = os.path.join(CHECKOUT_ROOT_DIR, "cinderx")
PYTHON_LIB_DIR = "cinderx/PythonLib"

MIN_GCC_VERSION = 13


def compute_package_version() -> str:
    """
    Compute a date-based version string.  Uses the UTC timezone for consistency.

    The returned format is YYYY.MM.DD.PP, where PP is a patch number.
    setuptools is going to normalize this anyway, so the two-character segments
    are going to get cut down if they have a leading zero.

    The patch number defaults to "00" but can be overridden with the
    CINDERX_VERSION_PATCH environment variable to allow multiple releases
    on the same day.

    When building from an extracted source distribution (sdist), the version
    is read from PKG-INFO to ensure consistency with the original release.
    """

    pkg_info_path = os.path.join(CHECKOUT_ROOT_DIR, "PKG-INFO")
    if os.path.exists(pkg_info_path):
        with open(pkg_info_path, "r") as f:
            for line in f:
                if line.startswith("Version:"):
                    version = line.split(":", 1)[1].strip()
                    print(f"Using version from PKG-INFO: {version}")
                    return version

    utc_now = datetime.datetime.now(datetime.timezone.utc)
    date_part = utc_now.strftime("%Y.%m.%d")
    patch = int(os.environ.get("CINDERX_VERSION_PATCH", "0"))
    return f"{date_part}.{patch:02}"


@lru_cache(maxsize=1)
def get_compiler() -> tuple[str, str]:
    """
    Prefers GCC if a new enough version is installed as this is what the
    cibuildwheel environment uses.

    Returns:
        A tuple of (c_compiler, cxx_compiler) paths.
    """
    cc_path = os.environ.get("CC")
    cxx_path = os.environ.get("CXX")
    if cc_path and cxx_path:
        print(f"Using CC and CXX from environment: {cc_path}, {cxx_path}")
        return (cc_path, cxx_path)

    gcc_path = shutil.which("gcc")
    gxx_path = shutil.which("g++")

    if gcc_path and gxx_path:
        try:
            result = subprocess.run(
                [gcc_path, "--version"],
                capture_output=True,
                text=True,
                check=True,
                timeout=5,
            )
            version_output = result.stdout

            # Parse GCC version from output like "gcc (GCC) 14.1.0"
            # The version is typically in the first line
            match = re.search(
                r"gcc.*?(\d+)\.(\d+)(?:\.(\d+))?", version_output, re.IGNORECASE
            )
            if match:
                major_version = int(match.group(1))
                print(f"Found GCC version {major_version}.{match.group(2)}")

                if major_version >= MIN_GCC_VERSION:
                    print(f"Using GCC: {gcc_path}, {gxx_path}")
                    return (gcc_path, gxx_path)
                else:
                    print(
                        f"GCC version {major_version} < {MIN_GCC_VERSION}, checking for Clang"
                    )
        except (subprocess.SubprocessError, subprocess.TimeoutExpired) as e:
            print(f"Failed to determine GCC version: {e}, checking for Clang")

    # Fall back to Clang
    clang_path = shutil.which("clang")
    clangxx_path = shutil.which("clang++")

    if clang_path and clangxx_path:
        print(f"Using Clang: {clang_path}, {clangxx_path}")
        return (clang_path, clangxx_path)

    raise RuntimeError("Cannot find suitable C/C++ compiler (tried gcc and clang)")


class PgoStage(Enum):
    DISABLED = 0
    GENERATE = 1
    USE = 3


def find_files(path: str, pred: Callable[[str], bool]) -> list[str]:
    result = []
    for root, dirs, files in os.walk(path):
        for f in files:
            if pred(f):
                # Files are returned relative to the top directory.
                abs_file = os.path.join(root, f)
                result.append(os.path.relpath(abs_file, path))
    return result


def find_native_sources(path: str) -> list[str]:
    return find_files(path, lambda f: f.endswith(".cpp") or f.endswith(".c"))


def compute_py_version() -> str:
    return f"{sys.version_info.major}.{sys.version_info.minor}"


class BuildCommand(build):
    # Don't use the setuptools default under "build/" as this clashes with
    # "build/fbcode_builder/" (auto-added to the OSS view of CinderX).
    def initialize_options(self):
        build.initialize_options(self)
        self.build_base = "scratch"

    def run(self) -> None:
        enable_pgo = os.environ.get("CINDERX_ENABLE_PGO", None) is not None

        if enable_pgo:
            self._run_with_pgo()
        else:
            super().run()

    def _run_with_pgo(self) -> None:
        def print_section(title: str) -> None:
            separator = "=" * 70
            print(f"\n{separator}")
            print(title)
            print(separator)

        cc, _ = get_compiler()
        is_clang = "clang" in cc

        print_section("PGO STAGE 1/3: Building with profile generation instrumentation")

        stage1_build_ext_cmd = self.get_finalized_command("build_ext")
        stage1_build_ext_cmd.cinderx_pgo_stage = PgoStage.GENERATE

        # Run normal build process (BuildPy + BuildExt)
        super().run()

        print_section("PGO STAGE 2/3: Running profiling workload")

        workload_env = os.environ.copy()

        if is_clang:
            clang_pgo_dir = os.path.join(self.build_temp, "pgo_data")
            os.makedirs(clang_pgo_dir, exist_ok=True)
            raw_profile_pattern = os.path.join(clang_pgo_dir, "code-%p.profraw")
            clang_merged_profile = os.path.join(clang_pgo_dir, "code.profdata")
            workload_env["LLVM_PROFILE_FILE"] = raw_profile_pattern

        # Add build output to PYTHONPATH so workload can import cinderx
        if "PYTHONPATH" in workload_env:
            workload_env["PYTHONPATH"] = (
                f"{self.build_lib}:{workload_env['PYTHONPATH']}"
            )
        else:
            workload_env["PYTHONPATH"] = self.build_lib

        # Uses the same default workload as CPython's PGO
        workload_cmd = [
            sys.executable,
            "-c",
            """
import cinderx

import sys
sys.argv.append("--pgo")

def main():
    # This import must not be in the module body as it will start the tests
    # running, and those using multiprocessing will fail because the initial
    # doesn't have "freeze support".
    import test.__main__

if __name__ == "__main__":
    main()
            """,
        ]

        print(f"Running workload with PYTHONPATH={workload_env['PYTHONPATH']}")
        workload_args = {
            "env": workload_env,
            "check": True,
        }
        if is_clang:
            workload_args["cwd"] = clang_pgo_dir
        subprocess.run(workload_cmd, **workload_args)

        if is_clang:
            print_section("PGO STAGE 2b: Merging profile data")

            llvm_profdata = shutil.which("llvm-profdata")
            if not llvm_profdata:
                raise RuntimeError("Cannot find llvm-profdata")
            profraw_files = glob.glob(os.path.join(clang_pgo_dir, "*.profraw"))

            if not profraw_files:
                raise RuntimeError(
                    "No profile data generated! Check that workload ran successfully."
                )

            print(f"Found {len(profraw_files)} profile files to merge")
            merge_cmd = [
                llvm_profdata,
                "merge",
                "-output=" + clang_merged_profile,
            ] + profraw_files
            subprocess.run(merge_cmd, check=True)
            print(f"Merged profile written to {clang_merged_profile}")

        print_section("PGO STAGE 3/3: Rebuilding with profile-guided optimizations")

        print("Cleaning build artifacts (keep profile data + python libs)...")

        # IMPORTANT: We need to preserve CMakeCache.txt to avoid re-running
        # compiler feature detection. For some reason asmjit seems to
        # consistently change its mind on the need for -fmerge-all-constants
        # after the instrumentation build, and this invalidates some profiling
        # data. However, we DO need to force CMake to reconfigure to pick up
        # the new PGO_STAGE=USE environment variable. We do this by removing
        # the CMakeFiles directory but keeping CMakeCache.txt.
        #
        # For GCC PGO, we must preserve .gcda files (profile data generated
        # during the workload run) as they are stored alongside object files
        # in CMakeFiles/. We selectively remove only what we need to force
        # reconfiguration.

        cmake_files = os.path.join(self.build_temp, "CMakeFiles")
        if os.path.exists(cmake_files):
            print(f"  Cleaning {cmake_files} (preserving .gcda/.gcno for GCC PGO)")
            for root, _dirs, files in os.walk(cmake_files, topdown=False):
                for f in files:
                    file_path = os.path.join(root, f)
                    # Preserve GCC PGO profiling files
                    if f.endswith((".gcda", ".gcno")):
                        continue
                    os.remove(file_path)
                # Remove empty directories, but only if they don't contain
                # preserved files (the topdown=False walk handles this)
                try:
                    os.rmdir(root)  # Only removes if empty
                except OSError:
                    pass  # Directory not empty, contains .gcda/.gcno files

        # Remove all object files and libraries to force rebuild
        for rm_root in (self.build_temp, self.build_lib):
            if os.path.exists(rm_root):
                for root, _dirs, files in os.walk(rm_root):
                    # Skip the pgo_data directory
                    if "pgo_data" in root:
                        continue
                    for f in files:
                        if f.endswith((".o", ".so", ".a")):
                            file_path = os.path.join(root, f)
                            print(f"  Removing {file_path}")
                            os.remove(file_path)

        print("Running rebuild with PGO optimizations...")

        # Create a new build_ext command instance with extensions list. The old
        # one from Stage 1 consumed its inputs and cannot be reused.
        stage3_build_ext_cmd = BuildExt(self.distribution)
        stage3_build_ext_cmd.build_lib = self.build_lib
        stage3_build_ext_cmd.build_temp = self.build_temp
        stage3_build_ext_cmd.inplace = False
        stage3_build_ext_cmd.force = True
        stage3_build_ext_cmd.cinderx_pgo_stage = PgoStage.USE
        if is_clang:
            stage3_build_ext_cmd.cinderx_pgo_profile_path = clang_merged_profile
        stage3_build_ext_cmd.finalize_options()
        stage3_build_ext_cmd.run()

        print_section("PGO BUILD COMPLETE!")


class BuildPy(build_py):
    def run(self) -> None:
        # I have no idea what is supposed to set this up, but if it isn't set
        # we'll get an AttributeError at runtime.
        if not hasattr(self.distribution, "namespace_packages"):
            self.distribution.namespace_packages = None

        super().run()

        # Copy opcodes/${PY_VERSION}/opcode.py to cinderx/opcode.py.
        py_version = compute_py_version()
        out_path = self.get_module_outfile(self.build_lib, ["cinderx"], "opcode")
        os.makedirs(os.path.dirname(out_path), exist_ok=True)
        self.copy_file(
            os.path.join(PYTHON_LIB_DIR, f"opcodes/{py_version}/opcode.py"),
            out_path,
            preserve_mode=False,
        )

        # For OSS builds always surface the CinderX import errors
        dev_build_file = os.path.join(self.get_package_dir("cinderx"), ".dev_build")
        print(f"Writing .dev_build file to {dev_build_file}")
        with open(dev_build_file, "w") as f:
            f.write("\n")


class CMakeExtension(Extension):
    """
    Subclass to indicate to BuildExt that this should be handled specially.
    """

    pass


class BuildExt(build_ext):
    def __init__(self, distribution: Distribution) -> None:
        super().__init__(distribution)
        self.cinderx_pgo_stage = PgoStage.DISABLED
        self.cinderx_pgo_profile_path: str | None = None

    def run(self) -> None:
        # Partition into CMake extensions and everything else.
        cmake_extensions = []
        other_extensions = []
        # pyre-ignore[16]: No pyre types for build_ext.
        for extension in self.extensions:
            if isinstance(extension, CMakeExtension):
                cmake_extensions.append(extension)
            else:
                other_extensions.append(extension)

        # Handle CMake specially, leave everything else to super().
        for extension in cmake_extensions:
            self._run_cmake(extension)

        # pyre-ignore[16]: No pyre types for build_ext.
        self.extensions = other_extensions
        super().run()

    def _run_cmake(self, extension: CMakeExtension) -> None:
        # pyre-ignore[16]: No pyre types for build_ext.
        build_dir = self.build_temp
        os.makedirs(build_dir, exist_ok=True)

        # pyre-ignore[16]: No pyre types for build_ext.
        extension_dir = os.path.abspath(self.get_ext_fullpath(extension.name))
        os.makedirs(extension_dir, exist_ok=True)

        cc, cxx = get_compiler()

        build_type = os.environ.get("CMAKE_BUILD_TYPE", "RelWithDebInfo")
        verbose_makefile = os.environ.get("CMAKE_VERBOSE_MAKEFILE", "OFF")
        cmake_args = [
            f"-DCMAKE_BUILD_TYPE={build_type}",
            f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={os.path.dirname(extension_dir)}",
            f"-DCMAKE_C_COMPILER={cc}",
            f"-DCMAKE_CXX_COMPILER={cxx}",
            f"-DCMAKE_VERBOSE_MAKEFILE:BOOL={verbose_makefile}",
        ]

        if self.cinderx_pgo_stage == PgoStage.GENERATE:
            cmake_args.append("-DENABLE_PGO_GENERATE=ON")
            cmake_args.append("-DENABLE_PGO_USE=OFF")
        elif self.cinderx_pgo_stage == PgoStage.USE:
            cmake_args.append("-DENABLE_PGO_GENERATE=OFF")
            cmake_args.append("-DENABLE_PGO_USE=ON")
            if self.cinderx_pgo_profile_path:
                cmake_args.append(f"-DPGO_PROFILE_FILE={self.cinderx_pgo_profile_path}")

        # LTO configuration
        enable_lto = os.environ.get("CINDERX_ENABLE_LTO", None)
        if enable_lto is not None:
            cmake_args.append("-DENABLE_LTO=ON")
            print("Building with LTO enabled (full LTO)")
        else:
            cmake_args.append("-DENABLE_LTO=OFF")

        options: dict[str, str] = {}

        def set_option(var: str, default: object) -> None:
            if type(default) == bool:
                default = int(default)
            if type(default) == int:
                default = str(default)
            if type(default) != str:
                raise ValueError(f"Not sure what to do with default value {default}")

            value = os.environ.get(var, default)
            options[var] = value

        # Python version is always the same as what's running setuptools.
        py_version = compute_py_version()
        options["PY_VERSION"] = py_version
        options["Python_ROOT_DIR"] = self._find_python()

        meta_python = "+meta" in sys.version
        linux = sys.platform == "linux"
        meta_312 = meta_python and py_version == "3.12"
        is_314plus = py_version == "3.14" or py_version == "3.15"

        set_option("META_PYTHON", meta_python)
        set_option("ENABLE_ADAPTIVE_STATIC_PYTHON", meta_312)
        set_option("ENABLE_DISASSEMBLER", False)
        set_option("ENABLE_ELF_READER", linux)
        set_option("ENABLE_EVAL_HOOK", meta_312)
        set_option("ENABLE_FUNC_EVENT_MODIFY_QUALNAME", meta_312)
        set_option("ENABLE_GENERATOR_AWAITER", meta_312)
        set_option("ENABLE_INTERPRETER_LOOP", meta_312 or is_314plus)
        set_option("ENABLE_LAZY_IMPORTS", meta_312)
        set_option("ENABLE_LIGHTWEIGHT_FRAMES", meta_312)
        set_option("ENABLE_PARALLEL_GC", meta_312)
        set_option("ENABLE_PEP523_HOOK", meta_312 or is_314plus)
        set_option("ENABLE_PERF_TRAMPOLINE", meta_312)
        set_option("ENABLE_SYMBOLIZER", linux)
        set_option("ENABLE_USDT", linux)

        for name, value in options.items():
            cmake_args.append(f"-D{name}={value}")

        build_args = [
            "--config",
            build_type,
            "--",
            "-j",
            str(os.cpu_count() or 1),
        ]

        # pyre-ignore[16]: No pyre types for build_ext.
        self.spawn(["cmake"] + cmake_args + ["-B", build_dir, CHECKOUT_ROOT_DIR])
        self.spawn(["cmake", "--build", build_dir] + build_args)

    def _find_python(self) -> str:
        # Normally this would use "data", but that goes to a temporary build directory
        # under uv.  Work off of the include directory instead.
        include_dir = sysconfig.get_path("include")
        return os.path.join(include_dir, "..", "..")


def main() -> None:
    # Native sources.
    native_sources = find_native_sources(SOURCE_DIR)

    setup(
        name="cinderx",
        description="High-performance Python runtime extension",
        url="https://www.github.com/facebookincubator/cinderx",
        version=compute_package_version(),
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
        ],
        ext_modules=[
            CMakeExtension(
                name="_cinderx",
                sources=native_sources,
            ),
        ],
        cmdclass={
            "build": BuildCommand,
            "build_py": BuildPy,
            "build_ext": BuildExt,
        },
        packages=find_packages(where=PYTHON_LIB_DIR, exclude=["test_cinderx*"]),
        package_dir={"": PYTHON_LIB_DIR},
        package_data={"cinderx": [".dev_build"]},
        python_requires=">=3.14,<3.16",
    )


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict
from typing import Callable


class Opcode:
    CMP_OP = (
        "<",
        "<=",
        "==",
        "!=",
        ">",
        ">=",
        "BAD",
    )
    CONTAINS_OP_ARGS = (
        "in",
        "not in",
    )
    IS_OP_ARGS = (
        "is",
        "is not",
    )

    HAVE_ARGUMENT = 90  # Opcodes from here have an argument:
    EXTENDED_ARG = 144
    EXTENDED_OPCODE = 126
    CODEUNIT_SIZE = 2

    def __init__(self) -> None:
        self.hasconst: set[str] = set()
        self.hasname: set[int] = set()
        self.hasjrel: set[int] = set()
        self.hasjabs: set[int] = set()
        self.haslocal: set[int] = set()
        self.hascompare: set[int] = set()
        self.hasfree: set[int] = set()
        self.shadowop: set[int] = set()

        self.opmap: dict[str, int] = {}
        self.opname: list[str] = ["<{!r}>".format(op) for op in range(256)]
        self.stack_effects: dict[str, object] = {}

        # New for 3.14, we need to know popped vs pushed for the load borrow analysis.
        self.popped: dict[str, int | Callable[[object], int]] = {}
        self.pushed: dict[str, int | Callable[[object], int]] = {}

    def stack_effect(self, opcode: int, oparg, jump: int) -> int:  # pyre-ignore[2]
        oparg_int = 0
        if opcode >= self.HAVE_ARGUMENT:
            if oparg is None:
                raise ValueError(
                    "stack_effect: opcode requires oparg but oparg was not specified"
                )
            oparg_int = int(oparg)
        elif oparg is not None:
            raise ValueError(
                "stack_effect: opcode does not permit oparg but oparg was specified"
            )
        # pyre-fixme[6]: For 1st argument expected `Optional[bool]` but got `int`.
        jump_int = {None: -1, True: 1, False: 0}.get(jump)
        if jump_int is None:
            raise ValueError("stack_effect: jump must be False, True or None")
        opname = self.opname[opcode]
        return self.stack_effect_raw(opname, oparg_int, jump_int)

    def stack_effect_raw(self, opname: str, oparg, jump: int) -> int:  # pyre-ignore[2]
        effect = self.stack_effects.get(opname)
        if effect is None:
            raise ValueError(
                f"Error, opcode {opname} was not found, please update opcode.stack_effects"
            )
        if isinstance(effect, int):
            return effect
        else:
            return effect(oparg, jump)  # pyre-ignore[29]

    def get_num_popped(self, opname: str, oparg: object) -> int:
        popped = self.popped.get(opname)
        if popped is None:
            raise ValueError(
                f"Error, opcode {opname} was not found, please update opcode.stack_effects"
            )
        if isinstance(popped, int):
            return popped
        else:
            return popped(oparg)

    def get_num_pushed(self, opname: str, oparg: object) -> int:
        pushed = self.pushed.get(opname)
        if pushed is None:
            raise ValueError(
                f"Error, opcode {opname} was not found, please update opcode.stack_effects"
            )
        if isinstance(pushed, int):
            return pushed
        else:
            return pushed(oparg)

    def def_op(self, name: str, op: int) -> None:
        # Fill in any missing space in the opname list for
        # opargs which are > 255.
        while op >= len(self.opname):
            self.opname.append("<{!r}>".format(op))
        self.opname[op] = name
        self.opmap[name] = op
        setattr(self, name, op)

    def name_op(self, name: str, op: int) -> None:
        self.def_op(name, op)
        self.hasname.add(op)

    def jrel_op(self, name: str, op: int) -> None:
        self.def_op(name, op)
        self.hasjrel.add(op)

    def jabs_op(self, name: str, op: int) -> None:
        self.def_op(name, op)
        self.hasjabs.add(op)

    def has_jump(self, op: int) -> bool:
        return op in self.hasjrel or op in self.hasjabs

    def shadow_op(self, name: str, op: int) -> None:
        self.def_op(name, op)
        self.shadowop.add(op)

    def remove_op(self, opname: str) -> None:
        op = self.opmap[opname]
        self.hasconst.discard(opname)
        self.hasname.discard(op)
        self.hasjrel.discard(op)
        self.hasjabs.discard(op)
        self.haslocal.discard(op)
        self.hascompare.discard(op)
        self.hasfree.discard(op)
        self.shadowop.discard(op)
        self.opmap.pop(opname)
        self.opname[op] = None  # pyre-ignore[6]
        self.stack_effects.pop(opname)
        delattr(self, opname)

    def copy(self) -> "Opcode":
        result = Opcode()
        result.hasconst = self.hasconst.copy()
        result.hasname = self.hasname.copy()
        result.hasjrel = self.hasjrel.copy()
        result.hasjabs = self.hasjabs.copy()
        result.haslocal = self.haslocal.copy()
        result.hascompare = self.hascompare.copy()
        result.hasfree = self.hasfree.copy()
        result.shadowop = self.shadowop.copy()
        result.opmap = self.opmap.copy()
        result.opname = self.opname.copy()
        result.stack_effects = self.stack_effects.copy()
        for name, op in self.opmap.items():
            setattr(result, name, op)
        return result
